import io
from typing import Tuple, List, Dict, Set
import cv2
import numpy as np
import pandas as pd
from PIL import Image
from segments import SegmentsClient, SegmentsDataset
from sklearn import metrics
from sklearn.model_selection import train_test_split
from sklearn.tree import DecisionTreeClassifier
from txp.ml.common.tasks import Policy
from txp.ml.common.tasks.slic_cv_tasks import SlicPatchImageProcessor
from txp.ml.common.tasks.task import Encoder, Task
import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


class SlicPatchRecognitionEncoder(Encoder):
    """Encoder definition for all the task implementations
    that uses the SlicPatch object recognition algorithm.
    """
    def __init__(
        self,
        number_of_segments: int,
        image_shape: Tuple,
        thermal_image_shape: Tuple,
        patch_x_y_image: Tuple,
        patch_x_y_thermal: Tuple,
        pandas_cols: List[str],
        schema: Dict
    ):
        super(SlicPatchRecognitionEncoder, self).__init__(schema)
        self._number_of_segments: int = number_of_segments
        self._image_shape: Tuple = image_shape
        self._thermal_image_shape: Tuple = thermal_image_shape
        self._patch_x_y_image: Tuple = patch_x_y_image
        self._patch_x_y_thermal: Tuple = patch_x_y_thermal
        self._slic_processor: SlicPatchImageProcessor = SlicPatchImageProcessor(
            self._number_of_segments
        )
        self._pandas_cols: List[str] = pandas_cols

    def build_training_dataset(self, tables, target, **kwargs) -> pd.DataFrame:
        """Builds the training dataset for the task.

        Args
          **kwargs:
            'segments_api_key' (str): The key to connect to segments API to download
                computer vision panoptic datasets
            'segments_dataset_id' (str): The dataset ID in segments hub.
            'segments_dataset_release' (str): The version release value for the
                dataset.
            'instance_ids' (List[int]): The list of instances ID's labels to
                recognize.
            'spot_id' (str): The spot name that helps to identify which images
                in the dataset correspond to the position to be trained.
            'image_type': The image type to process: 'image' or 'thermographic'
            'true_categories' (List[int]): The categories IDs present in an labeled image that
                indicates the presence of the object.
            'dataset' (pd.DataFrame): a training pandas DataFrame. If provided, the
                nothing will be downloaded from Segments.ai server.
        """
        if 'dataset' not in kwargs:
            segs_client = self._get_segments_client(
                kwargs['segments_api_key']
            )

            segments_ds = self._get_segments_dataset(
                kwargs['segments_dataset_id'],
                kwargs['segments_dataset_release'],
                segs_client
            )

            instance_ids = set(kwargs['instance_ids'])
            spot_id = kwargs['spot_id']
            image_type = kwargs['image_type']
            true_categories = kwargs['true_categories']

            df: pd.DataFrame = self._map_segments_to_pandas(
                segments_ds, segs_client, image_type, instance_ids, spot_id, true_categories
            )

        else:
            df = kwargs['dataset']

        self.dataset = df
        self.target = 'label'

        return df

    def transform_input(self, signals_completed: dict) -> pd.DataFrame:
        vector = []

        for key_edge_logical_id, value_perceptions in self.schema.items():
            for key_percept, sch_values in value_perceptions.items():
                for table_id in sch_values:
                    if table_id:
                        image = [
                            int(x) for x in
                            signals_completed[key_edge_logical_id][key_percept][table_id][0]["data"][0]
                        ]
                        img_np_array = self._get_image_array(image)
                        log.info(f"{self.__class__.__name__} Prediction "
                                 f"with image shape: {img_np_array.shape}")
                        print(f"{self.__class__.__name__} Prediction "
                                 f"with image shape: {img_np_array.shape}")
                        vector = self._get_slic_vector(img_np_array, key_percept, vector)

        dataset = pd.DataFrame()
        dataset = dataset.append([vector], ignore_index=True)
        return dataset

    def _map_segments_to_pandas(
        self,
        segments_ds: SegmentsDataset,
        segments_client: SegmentsClient,
        image_type: str,
        instances_ids: Set[str],
        spot_id: str,
        true_categories: Tuple
    ) -> pd.DataFrame:
        """This method allows to parse a Segments.ai dataset into a Pandas
        Dataframe that can be consumed by the Task object model.

        Note: For our first usage of Segments.ai, we labelled all the images on the same
            Dataset. In consequence, we needed to add descriptive information at labeling time
            that allowed us to know which spot and instance IDs was being processed.

        TODO: we should revisit that conceptual labeling values, and this code, If we decide
            to move forward using Segments.ai

        Args:
            segments_ds: The segments dataset object.
                https://docs.segments.ai/reference/python-sdk#datasets

            segments_client: A SegmentsClient completely authenticated.
                https://docs.segments.ai/reference/python-sdk#setup

            image_type (str): The image type to process. { 'image', 'thermographic' }

            instances_ids (List[str]): The list of instances ids to recognize.

            spot_id(str): The value of the spot ID to train.

            true_categories: The categories IDs present in an labeled image that
                indicates the presence of the object.
        """
        df_dicts = []
        normal_res = (self._image_shape[0], self._image_shape[1])
        thermographic_res = (self._thermal_image_shape[0], self._thermal_image_shape[1])
        patch_to_use = self._patch_x_y_image if image_type == 'image' else self._patch_x_y_thermal
        for sample in segments_ds:
            if image_type == 'image':
                if sample['image'].size != normal_res:
                    continue

            else:
                if sample['image'].size != thermographic_res:
                    continue

            label = segments_client.get_label(sample['uuid'])

            image_attributes = label['attributes']['image_attributes']

            if image_attributes.get('spot_id', None) != spot_id:
                continue

            sample_cats = (set(map(
                lambda ann: ann['category_id'],
                label['attributes']['annotations']
            )))

            has_category = not sample_cats.isdisjoint(true_categories)

            # compute beer instance
            beer_instance = None
            for ann in label['attributes']['annotations']:
                if ann.get('attributes', None) is not None:
                    if ann['attributes'].get('instance_id', None) in instances_ids:
                        beer_instance = ann['id']
                        has_category = True
                        break
                    else:
                        beer_instance = -1
                        has_category = False

            bitmap_label = np.asarray(sample['segmentation_bitmap'].convert('L'))
            if beer_instance and beer_instance > 0:
                x, y, w, h = self._get_patch_coordinates_to_sample(bitmap_label, beer_instance)
            else:
                x, y, w, h = 0, 0, 0, 0

            df_dicts.append(
                {
                    self._pandas_cols[0]: self._slic_processor.process_image_patches(
                        np.asarray(sample['image']),
                        patch_to_use[0],
                        patch_to_use[1]
                    ),
                    self._pandas_cols[1]: np.asarray(sample['segmentation_bitmap'].convert('L')),
                    self._pandas_cols[2]: (w, h),
                    self._pandas_cols[3]: beer_instance,
                    self._pandas_cols[4]: int(has_category)
                }
            )

            df = pd.DataFrame(df_dicts, columns=self._pandas_cols)

        return df

    def _get_slic_vector(self, img_array: np.ndarray, key_percept: str, carry_array: List) -> List[float]:
        """
            TODO: Why are we using `extend` ? Ask.
        """
        if key_percept == "Image":
            if len(carry_array) > 0:
                carry_array.extend(self._slic_processor.process_image_patches(
                    img_array,
                    self._patch_x_y_image[0],
                    self._patch_x_y_image[1]
                ))
            else:
                carry_array = self._slic_processor.process_image_patches(
                    img_array,
                    self._patch_x_y_image[0],
                    self._patch_x_y_image[1]
                )
        elif key_percept == "ThermalImage":
            if len(carry_array) > 0:
                carry_array.extend(
                    self._slic_processor.process_image_patches(
                        img_array,
                        self._patch_x_y_image[0],
                        self._patch_x_y_image[1]
                    )
                )
            else:
                carry_array = self._slic_processor.process_image_patches(
                    img_array,
                    self._patch_x_y_image[0],
                    self._patch_x_y_image[1]
                )

        return carry_array

    @staticmethod
    def _get_image_array(image_data: List) -> np.ndarray:
        image = np.frombuffer(bytes(image_data), dtype=np.uint8)
        image_b = bytes(image)
        image_pil = Image.open(io.BytesIO(image_b))
        img_np_array = np.array(image_pil)
        return img_np_array

    # ======================== build_training_dataset section =======================
    @staticmethod
    def _get_segments_client(api_key) -> SegmentsClient:
        client = SegmentsClient(api_key)
        return client

    @staticmethod
    def _get_segments_dataset(dataset_id, dataset_release, client) -> SegmentsDataset:
        release = client.get_release(dataset_id, dataset_release)
        dataset = SegmentsDataset(release, filter_by=["reviewed"])  # Always filter reviewed images
        return dataset

    @staticmethod
    def _get_instance_bitmap_mask(semantic_label_bitmap, instance_id: id) -> np.ndarray:
        """Receives a labeled semantic image bitmap, and mask all the pixels
        that don't belong to the instance_id"""
        mask = np.copy(semantic_label_bitmap)

        mask[mask != instance_id] = 0
        mask[mask == instance_id] = 255

        return mask

    @staticmethod
    def _get_patch_coordinates_to_sample(semantic_label_bitmap, instance_id, error=10) -> Tuple:
        """given a semantic_label image, obtains the rectangle coordinates to patch on
        based on the labels"""
        bitmap_interest_mask = SlicPatchRecognitionEncoder._get_instance_bitmap_mask(
            semantic_label_bitmap, instance_id
        )
        x, y, w, h = cv2.boundingRect(bitmap_interest_mask)
        return x, y, w, h


class SlicPatchRecognitionDecisionTreePolicy(Policy):
    def __init__(self, policy_file: str = None):
        super().__init__(policy_file)
        if self.clf is None:
            self.clf = DecisionTreeClassifier()

    def name(self):
        return self.__class__.name()

    def train(
            self,
            dataset: pd.DataFrame,
            target: str = 'label',
            test_size: float = 0.3,
    ):
        """
        Args:
            dataset: the training dataset MUST have the columns:
                `image`: The features vector of the sample.
                `target`: The label value for the sample (1 or 0)

            target (str): The column name for the targets
            test_size ( float ): The percentage of test data to perform the dataframe split
        """
        x = dataset['image'].values.tolist()
        y = dataset[target].values.tolist()
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=test_size, random_state=0)

        self.clf = self.clf.fit(x_train, y_train)
        y_prediction = self.clf.predict(x_test)

        self.accuracy = metrics.accuracy_score(y_test, y_prediction)
        self.trained = True

    def predict(self, tensor):
        if self.is_trained():
            return self.clf.predict(tensor)
        else:
            return None


class SlicPatchRecognitionTask(Task):
    """General definition for all the task implementations
    that uses the SlicPatch object recognition algorithm.
    """
    def __init__(
        self,
        task_definition=None,
        credentials_str=None,
    ):
        super().__init__(
            self._build_encoder(task_definition.schema_, task_definition.parameters),
            self._build_policy(),
            task_definition,
            credentials_str
        )

        if self.task_definition.task_data and credentials_str is not None:
            self.load_from_bucket(self.task_definition.task_data)

            if task_definition.schema_:
                self.encoder.set_schema(task_definition.schema_)

        self.ready = True

        log.info(f"{self.__class__.__name__} instance is ready with file: {self.task_definition.task_data}")
        print(f"{self.__class__.__name__} instance is ready with file: {self.task_definition.task_data}")

    def _build_encoder(
        self, schema: Dict, parameters: Dict
    ) -> SlicPatchRecognitionEncoder:
        """Return a new instance reference to a SlicPatchRecognitionEncoder."""
        return SlicPatchRecognitionEncoder(
            parameters['number_of_segments'],
            parameters['image_shape'],
            parameters['thermal_image_shape'],
            parameters['patch_x_y_image'],
            parameters['patch_x_y_thermal'],
            parameters['pandas_cols'],
            schema
        )

    def _build_policy(self) -> Policy:
        """Returns the instance reference to the appropriated Policy type."""
        return SlicPatchRecognitionDecisionTreePolicy()

    def predict(self, sampling_window_rows) -> str:
        """Returns the prediction results for the given sampling windows
        generated input vectors.

        Args:
            sampling_window_rows: filled Firestore dictionary, this dict
                contains all the rows present in bigquery tables.

        Returns:
            Appropriate enumerated label.
        """
        print(f"{self.__class__.__name__} prediction request.")
        prediction_dataset = self.encoder.transform_input(sampling_window_rows)
        return self.task_definition.label_def.int_to_label[
            self.policy.predict(prediction_dataset)[0]
        ]

    def train(self, dataset: pd.DataFrame) -> None:
        self.policy.train(dataset, self.encoder.target)



"""
This module implements the Computer Vision Analytics View for the application.

NOTE: This is just a dummy implementation of a MainView in order to shown the
mechanism provided by the `core_components` package.
"""
import pytz
from txp.common.config import settings
from pylab import *
import pandas as pd
import plotly.express as px
import txp.web.views.analytics.computer_vision_analytics as computer_vision_utils
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from txp.cloud.controllers.authentication import AuthenticationInfo

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


def change_utc_timezone(timestamp):
    utc = pytz.timezone("UTC")
    timezone = pytz.timezone("America/Mexico_City")
    date_time = pd.to_datetime(timestamp)
    localized_timestamp = utc.localize(date_time)
    new_timezone = localized_timestamp.astimezone(timezone)
    return str(new_timezone)


def plot_asset_temperatures(name, temperature_values, temperature_dates):
    """Creates a scatter plot with only temperature data
    https://plotly.com/python-api-reference/generated/plotly.graph_objects.Scatter.html
    """
    if not temperature_values is None:
        strtimes = temperature_dates

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=strtimes[::-1],
                y=temperature_values[::-1],
                marker=dict(
                    cmax=12,
                    cmin=-2,
                    size=12,
                    color=temperature_values[::-1],
                    colorscale="portland",
                    colorbar=dict(
                        title="Temperatura",
                        tickmode="array",
                        tickvals=[-2, 12],
                        ticktext=["low", "high"],
                        ticks="outside",
                    ),
                ),
                mode="markers",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=strtimes[::-1],
                y=temperature_values[::-1],
                mode="lines",
                line=dict(color="grey", width=2, dash="dash"),
            )
        )

        fig.update_layout(width=900, margin=dict(l=0, r=0, b=0, t=0), showlegend=False)
        fig.update_yaxes(range=[-2,12])
        return fig
    else:
        return None


def _get_temperatures(signals_from_interval):
    if not signals_from_interval:
        logging.info(f"No hay data disponible")
        return None, None
    else:
        temperatures = []
        dates = []
        start = time.perf_counter()
        for signal in signals_from_interval:
            dates.append(change_utc_timezone(signal["observation_timestamp"]))
            for dimension in signal["data"]:
                temperatures.append(float(dimension["values"][0]))
        end = time.perf_counter()
        logging.info(
            f"La descarga total de las {len(temperatures)} imagenes tardó {(end - start)}"
        )
        return temperatures, dates


def get_complex_bar_animation(
    user, images, name, step_names, temperatures_values, current_machine, labels
):
    """Creates a subplot animation with images and temperature bar chart
     in the same row.
    https://plotly.com/python-api-reference/generated/plotly.subplots.make_subplots.html
    https://plotly.com/python-api-reference/generated/plotly.graph_objects.Bar.html
    https://plotly.github.io/plotly.py-docs/generated/plotly.graph_objects.Image.html
    Its called complex because it displays data of different perceptions.
    """
    if temperatures_values is None:
        fig = make_subplots(rows=1, cols=1)
    else:
        fig = make_subplots(
            rows=1,
            cols=2,
            column_widths=[0.90, 0.10],
            horizontal_spacing=0.15,
            subplot_titles=(f"{name}  \n", "Temperaturas  \n"),
        )

    imgs = []
    dates = []
    annotations = []
    temperatures = []
    traces_added = False
    img_width = 600
    img_height = 600
    frames = []
    spots = [f"{current_machine}"] * len(temperatures_values)

    for i, (img, date, temp, label) in enumerate(list(
        zip(images, step_names, temperatures_values, labels)
    )):
        try:
            plot_fig = px.imshow(np.asarray(img))
            imgs.append(plot_fig)
            if not traces_added:
                fig.add_trace(imgs[0].data[0], row=1, col=1)
                if temperatures_values is not None:
                    fig.add_trace(
                        go.Bar(
                            x=np.asarray(spots[0]),
                            y=np.asarray(temperatures_values[0]),
                            text=temperatures_values[0],
                            textposition="auto",
                        ),
                        row=1,
                        col=2,
                    )
                traces_added = True
            dates.append(date)
            annotations.append(label)
            temperatures.append(temp)

            if user == AuthenticationInfo.UserRole.Admin:
                if label is not None:
                    layout_i = go.Layout(
                        annotations=[
                            go.layout.Annotation(
                                x=0.5,
                                y=0.9,
                                text=label,
                                xref="paper",
                                yref="paper",
                                showarrow=False,
                                font_size=20,
                                font_color="black",
                                bgcolor="white",
                                opacity=0.85,
                            )
                        ]
                    )
                else:
                    layout_i = go.Layout(
                        annotations=[
                            go.layout.Annotation(
                                x=300,
                                y=50,
                                text="Sin etiqueta",
                                showarrow=False,
                                font_size=20,
                                font_color="black",
                                bgcolor="white",
                                opacity=0.85,
                            )
                        ]
                    )
            else:
                layout_i = fig.layout

            if temperatures is not None:
                frames.append(
                    dict(
                        name=str(date),
                        layout=layout_i,
                        data=[
                            plot_fig.data[0],
                            go.Bar(
                                y=np.asarray(temperatures[i]),
                                text=f"{temperatures[i]}°",
                            ),
                        ],
                        traces=[0, 1],
                    )
                )

            else:
                frames.append(
                    dict(
                        name=str(date),
                        layout=layout_i,
                        data=[
                            plot_fig.data[0],
                        ],
                        traces=[0],
                    )
                )
        except Exception as e:
            log.error(f"Error catched: {e}. Probably a truncated image was found.")

    fig.frames = frames
    fig.update_layout(
        sliders=[
            {
                "active": 0,
                "x": 0.1,
                "y": 0,
                "len": 0.9,
                "pad": {"b": 10, "t": 40, "r": 10, "l": 10},
                "steps": [
                    {
                        "args": [
                            [fr.name],
                            {
                                "frame": {"duration": 0, "redraw": True},
                                "mode": "immediate",
                                "fromcurrent": True,
                            },
                        ],
                        "label": fr.name,
                        "method": "animate",
                    }
                    for fr in fig.frames
                ],
            }
        ],
    )

    fig.update_layout(
        updatemenus=[
            {
                "buttons": [
                    {
                        "args": [
                            None,
                            {
                                "frame": {"duration": 700, "redraw": True},
                                "fromcurrent": True,
                                "mode": "inmediate",
                                "transition": {
                                    "duration": 0,
                                    "easing": "quadratic-in-out",
                                },
                            },
                        ],
                        "label": "►►",
                        "method": "animate",
                    },
                {
                    "args": [
                        None,
                        {
                            "frame": {"duration": 1200, "redraw": True},
                            "fromcurrent": True,
                            "transition": {
                                "duration": 200,
                                "easing": "quadratic-in-out",
                            },
                        },
                    ],
                    "label": "▶",
                    "method": "animate",
                },
                    {
                        "args": [
                            [None],
                            {
                                "frame": {"duration": 0, "redraw": True},
                                "mode": "immediate",
                                "transition": {"duration": 0},
                            },
                        ],
                        "label": "■",
                        "method": "animate",
                    },
                ],
                "direction": "left",
                "pad": {"r": 10, "t": 15},
                "showactive": False,
                "type": "buttons",
                "x": 0.1,
                "xanchor": "right",
                "y": 0,
                "yanchor": "top",
            }
        ]
    )
    fig.update_layout(
        margin=dict(l=0, r=100, b=50, t=60), height=img_height, width=img_width
    )

    # Configure axes
    fig.update_xaxes(visible=False, row=1, col=1)
    fig.update_yaxes(
        visible=False,
        # the scaleanchor attribute ensures that the aspect ratio stays constant
        scaleanchor="x",
        row=1,
        col=1,
    )
    fig.update_yaxes(range=[min(temperatures) - 2, max(temperatures) + 2], row=1, col=2)
    fig.update_xaxes(visible=False, row=1, col=2)

    return fig


def get_images_and_temperature_analysis(
    user,
    normal_data,
    thermographic_data,
    range_time,
    data_format,
    images_data_format,
    timages_data_format,
    animation_data_format,
    temperatures_values,
    temperatures_dates,
    current_machine,
    images_charts,
    temperature_charts,
):
    temperature_data_format = "Temperaturas"
    (
        normal_images,
        normal_dates,
        normal_labels,
    ) = computer_vision_utils.get_images_and_dates(normal_data)
    (
        thermographic_images,
        thermographic_dates,
        thermographic_labels,
    ) = computer_vision_utils.get_images_and_dates(thermographic_data)

    if normal_images:
        images_charts.normal_imgs_chat = get_complex_bar_animation(
            user,
            normal_images,
            images_data_format,
            normal_dates,
            temperatures_values,
            current_machine,
            normal_labels,
        )

    else:
        images_charts.normal_imgs_chat = None

    if thermographic_images:
        images_charts.thermographic_imgs_datacharts = get_complex_bar_animation(
            user,
            thermographic_images,
            timages_data_format,
            thermographic_dates,
            temperatures_values,
            current_machine,
            thermographic_labels,
        )

    else:
        images_charts.thermographic_imgs_datacharts = None

    if temperatures_values:
        temperature_charts.temp_chart = plot_asset_temperatures(
            temperature_data_format, temperatures_values, temperatures_dates
        )
    else:
        temperature_charts.temp_chart = None



from scipy.fft import fft, fftfreq
import numpy as np
from scipy.fft import fft
import matplotlib.pyplot as plt


class FFTPlotter:
    """Fft plotter util.

    Defines a plotter utility class for plotting fft signals.

    Args:
        signal_amplitude_axis: Amplitude axis of the signal, if this argument is None
        then signal_frequency_amplitude_axis has to be given.

        signal_frequency_amplitude_axis: Precalculated fft, if fft has to be calculated within
        the plotter, then this Argument has to be set as None.

        sampling_frequency: frequency of sampling of the signal.
    """

    def __init__(self, signal_amplitude_axis: np.ndarray, signal_frequency_amplitude_axis,
                 sampling_frequency: float):

        if signal_amplitude_axis is None and signal_frequency_amplitude_axis is None:
            raise Exception("signal_amplitude_axis and signal_frequency_amplitude_axis both None")

        self.signal_amplitude_axis = signal_amplitude_axis
        self.signal_frequency_amplitude_axis = signal_frequency_amplitude_axis

        if self.signal_frequency_amplitude_axis is None:
            self.signal_frequency_amplitude_axis = fft(self.signal_amplitude_axis)

        self.sample_count = len(self.signal_frequency_amplitude_axis)

        assert self.sample_count == len(self.signal_frequency_amplitude_axis)
        self.sampling_frequency = sampling_frequency
        self.frequency_axis = fftfreq(self.sample_count, 1 / self.sampling_frequency)
        self.frequency_axis_indices = self.frequency_axis > 0
        self.frequency_axis = self.frequency_axis[self.frequency_axis_indices]
        self.signal_frequency_amplitude_axis = self.signal_frequency_amplitude_axis[self.frequency_axis_indices]

    def plot(self, title, semilogy=False, x_scale=None):
        """Plots fft.
        Args:
            title: tittle for the plot.
            semilogy: If fft needs to be plotted in semi logarithmic scale this argument must be set as True.
            x_scale: x ticks for plotting
        Returns:
            Plotting figure
        """
        fig = plt.figure(figsize=(23, 10))
        fig.suptitle(title, fontsize=20)
        if x_scale is not None:
            plt.xticks(np.arange(0, max(self.frequency_axis) + x_scale, x_scale))
        if semilogy:
            plt.semilogy(self.frequency_axis, 2.0 / self.sample_count * np.abs(self.signal_frequency_amplitude_axis),
                         '-b')
        else:
            plt.plot(self.frequency_axis, 2.0 / self.sample_count * np.abs(self.signal_frequency_amplitude_axis), '-b')
        plt.legend(['FFT'])
        return fig


class PSDPlotter:

    def __init__(self, frequency, psd):
        self.frequency = frequency
        self.psd = psd

    def plot(self, title, semilogy=False, x_scale=None):

        fig = plt.figure(figsize=(23, 10))
        fig.suptitle(title, fontsize=20)
        if x_scale is not None:
            plt.xticks(np.arange(0, max(self.frequency) + x_scale, x_scale))
        if semilogy:
            plt.semilogy(self.frequency, self.psd, '-r')
        else:
            plt.plot(self.frequency, self.psd, '-r')
        plt.legend(['PSD'])
        return fig



#!/usr/bin/python

from SmartMeshSDK.ApiException import CommandError

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('ApiDefinition')
log.setLevel(logging.ERROR)
log.addHandler(NullHandler())

class FieldFormats(object):
    '''
    \brief Enumeration of possible field formats.
    '''
    STRING              = 'string'
    BOOL                = 'bool'
    INT                 = 'int'
    INTS                = 'ints'
    HEXDATA             = 'hex'
    HEXDATA_VL          = 'hex_vl'      # variable length
    FLOAT               = 'float'
    ARRAY               = 'array'

class FieldOptions(object):
    '''
    \brief Possible options for a command field
    '''
    
    optionName     = None
    validOptions   = None
    optionDescs    = None
    
    def __init__(self,optionsDef,fieldOptions,fieldName):
        self.optionDescs = []
        if   not optionsDef:
            self.validOptions = None
        else:
            if optionsDef==True:
                self.optionName = fieldName
            else:
                self.optionName = optionsDef
            self.validOptions = []
            for i in fieldOptions[self.optionName]:
               self.validOptions.append(i[0])
               self.optionDescs.append(i[1])
    
    def isValidValue(self,val):
        if (
            (not self.validOptions)   or
            (val in self.validOptions)
           ):
            return True
        return False
    
    def valueToDesc(self,val):
        if self.validOptions:
            for i in range(len(self.validOptions)):
                if self.validOptions[i]==val:
                    return self.optionDescs[i]
        raise CommandError(CommandError.VALUE_NOT_IN_OPTIONS,
                           'option='+str(self.optionName)+' val='+str(val))

class Field(object):
    '''
    \brief Object representing one field of a command.
    '''
    
    def __init__(self,fieldDef,fieldOptions):
        self.name       = fieldDef[0]
        self.format     = fieldDef[1]
        self.length     = fieldDef[2]
        self.options    = FieldOptions(fieldDef[3],fieldOptions,self.name)
    
    def isValidValue(self,val):
        # check format and length
        if   self.format==FieldFormats.STRING:
            if ( (type(val) not in [string,unicodedata]) or
                 len(val)>self.length
               ):
                return False
        elif self.format==FieldFormats.BOOL:
            if type(val)!=bool:
                return False
        elif self.format==FieldFormats.INT:
            if ( (type(val)!=int and  type(val)!=long)  or
                 val>pow(2,8*self.length)
               ):
                return False
        elif self.format==FieldFormats.INTS:
            if ( (type(val)!=int and  type(val)!=long)  or
                 val>(pow(2,8*self.length)/2) or
                 val<(-pow(2,8*self.length)/2)
               ):
                return False
        elif self.format==FieldFormats.HEXDATA:
            if type(val)!=list and type(val)!=tuple:
                return False
            if self.length and len(val)>self.length:
                return False
            for i in val:
                if type(i)!=int:
                    return False
                if i>=pow(2,8):
                    return False
        elif self.format==FieldFormats.FLOAT:
            if ( (type(val)!=int and type(val)!=float) ):
                return False
        else:
            raise SystemError('unknown field format='+self.format)
        
        # check options
        if self.options.isValidValue(val)==False:
            return False
        
        return True

class ApiDefinition(object):
    '''
    \ingroup ApiDefinition
    
    \brief Base class for all API definitions objects.
    '''
    
    RC           = 'RC'
    SUBID1       = '_subId1'
    SUBID2       = '_subId2'
    RESERVED     = [SUBID1,SUBID2,'magic']
    OPTIONAL     = []
    COMMAND      = 'command'
    NOTIFICATION = 'notification'
    RC_OK        = 0
    
    #======================== id and name =====================================
    def __init__(self, array2scalar = True) :
        if array2scalar :
            self._array2scalar(self.commands)
            self._array2scalar(self.notifications)
    
    def _array2scalar(self, defs) :
        '''
        \brief Convert ARRAY to list of scalars
        '''
        for fields in defs:
            if 'subCommands' in fields :
                self._array2scalar(fields['subCommands'])
            if 'response' in fields and 'FIELDS' in fields['response'] :
                arrays = [f for f in fields['response']['FIELDS'] if f[1] == FieldFormats.ARRAY]
                for array in arrays :
                    scalars = []
                    for n in range(array[2]) :
                        for item in array[3] :
                            name = '{}_{}'.format(item[0], n+1)
                            scalars.append([name] + item[1:])
                    fields['response']['FIELDS'].remove(array)
                    fields['response']['FIELDS'] += scalars
    
    def idToName(self,type,id):
        '''
        \brief Translate a command or notification ID into a command name.
       
        \exception CommandError.INVALID_COMMAND Command does
                   not exist
        \returns The command name.
        '''
        list = self._getList(type)
        for item in list:
            if item['id']==id:
               return item['name']
        raise CommandError(CommandError.INVALID_COMMAND,
                           'id=%s' % str(id))
    
    def nameToId(self,type,nameArray):
        '''
        \brief Translate a command or notification name into a command ID.
       
        \exception CommandError.INVALID_COMMAND Command does
                   not exist
        \returns The command ID.
        '''
        list = self._getList(type)
        for item in list:
            if item['name']==nameArray[0]:
               return item['id']
        raise CommandError(CommandError.INVALID_COMMAND,
                                        nameArray[0])
    
    def rcToLabel(self,rc):
        '''
        \brief Translate a return code (RC) into its label, i.e. 'RC_OK' for 0x00.
        
        \param rc A return code, an int.
        
        \exception CommandError If RC does not exist.
        
        \returns The label for this RC, a string.
        '''
        
        # get the RC description
        rcLabel       = None
        for r in self.fieldOptions[self.RC]:
            if r[0]==rc:
                rcLabel           = r[1]
                break
        if not rcLabel:
            raise CommandError(CommandError.VALUE_NOT_IN_OPTIONS,
                               'rc={0} does not exist'.format(rc))
        
        return rcLabel
    
    def rcToDescription(self,rc,nameArray):
        '''
        \brief Translate a return code (RC) into a description.
       
        If this RC is described in the API definition for thise nameArray, then
        that description is returned. If not, the generic description of that
        RC is returned, preceeded by the string "[generic]".
        
        \param rc A return code, an int.
        
        \exception CommandError.VALUE_NOT_IN_OPTIONS if rc not in generic RC's.
        \returns The description for this RC, a string.
        '''
        
        returnVal = ''
        
        # get the RC description
        rcLabel       = None
        rcGenericDesc = None
        for r in self.fieldOptions[self.RC]:
            if r[0]==rc:
                rcLabel           = r[1]
                rcGenericDesc     = r[2]
                break
        if not rcLabel:
            raise CommandError(CommandError.VALUE_NOT_IN_OPTIONS,
                               'rc={0} does not exist'.format(rc))
        
        # retrieve rcDescription
        definition = self.getDefinition(self.COMMAND,nameArray)
        try:
            returnVal = definition['responseCodes'][rcLabel]
        except KeyError:
            returnVal = '[generic] {0}'.format(rcGenericDesc)
        
        return returnVal
    
    def getIds(self,type):
        '''
        \brief Get the list of command IDs this API defines
       
        \returns A array of numbers, each numbers representing a command ID
        '''
        list = self._getList(type)
        return [item['id'] for item in list]
    
    def getNames(self,type,nameArray=None):
        '''
        \brief Get the list of (sub)command names this API defines
       
        \param type         Type of definition to be looked up 
                            Supported values are: COMMAND or NOTIFICATION
        
        \param nameArray    Optional, used only when accessing names of
                            subcommands. Specifies the name of the command we
                            want the subcommand names for.
       
        \returns A array of strings, each string representing the name of a
                 command
        '''
        list = self._getList(type)
        definition  = None
        if nameArray:
            definition,list = self._commandIterator(nameArray,list)
        if not list:
            raise CommandError(CommandError.INVALID_COMMAND,
                                            '.'.join(nameArray)) 
        return [command['name'] for command in list]
    
    def getDefinition(self,type,nameArray):
        '''
        \brief Get the complete definition of a (sub)command, from its name.
        
        \param type           Type of definition to be looked up 
                              Supported values are: COMMAND or NOTIFICATION
        
        \param nameArray   An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
       
        \exception CommandError(INVALID_COMMAND) The (sub)command
                   does not exist.
        \returns The definition of a (sub)command, represented as a dictionary.
        '''
        list = self._getList(type)
        definition  = None
        definition,list = self._commandIterator(nameArray,list)
        return definition
        
    def getDescription(self,type,nameArray):
        '''
        \brief Get the description of a command.
       
        \param type           Type of definition to be looked up 
                              Supported values are: COMMAND or NOTIFICATION
        
        \param nameArray      An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
       
        \exception CommandError(INVALID_COMMAND) The (sub)command
                   does not exist.
        \returns The description of a (sub)command, represented as a dictionary.
                 '' if no description
        '''
        definition = self.getDefinition(type,nameArray)
        return definition['description']
    
    def hasSubcommands(self,type,nameArray):
        return 'subCommands' in self.getDefinition(type,nameArray)
    
    def subcommandIdToName(self,type,nameArray,id):
        subcommands = self.getSubcommands(type,nameArray)
        for subcommand in subcommands:
            if subcommand['id']==id:
                return subcommand['name']
        raise CommandError(CommandError.UNKNOWN_SUBCOMMAND,
                                            str(id))
    
    def subcommandNameToId(self,type,nameArray,name):
        subcommands = self.getSubcommands(type,nameArray)
        for subcommand in subcommands:
            if subcommand['name']==name:
                return subcommand['id']
        raise CommandError(CommandError.UNKNOWN_SUBCOMMAND,
                                            str(name))
    
    def getSubcommands(self,type,nameArray):
        definition = self.getDefinition(type,nameArray)
        if not 'subCommands' in definition:
            raise CommandError(CommandError.NO_SUBCOMMANDS,
                                            '.'.join(nameArray))
        return definition['subCommands']
    
    def _getList(self,type):
        if   type==self.COMMAND:
            list = self.commands
        elif type==self.NOTIFICATION:
            list = self.notifications
        else:
            raise ValueError("type="+str(type)+" unsupported")
        return list
    
    def _commandIterator(self,nameArray,list):
        for commandName in nameArray:
            if not list:
                raise CommandError(CommandError.INVALID_COMMAND,
                                            '.'.join(nameArray))
            found = False
            for elem in list:
                if elem['name']==commandName:
                    found       = True
                    definition  = elem
                    if 'subCommands' in elem:
                        list = definition['subCommands']
                    else:
                        list = None
                    break
            if found==False:
                raise CommandError(CommandError.INVALID_COMMAND,
                                            '.'.join(nameArray))
        return definition,list
    
    def getRequestFieldNames(self,commandArray):
        '''
        \brief Get the request fields of a (sub)command, from its name.
        
        \param commandArray   An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
       
        \exception CommandError(NO_REQUEST) The (sub)command has
                   no request fields.
        \returns The list of request fields.
        '''
        fields = self.getRequestFields(commandArray)
        return [field.name for field in fields]
    
    def getRequestFieldFormat(self,commandArray,fieldName):
        return self.getRequestField(commandArray,fieldName).format
        
    def getRequestFieldLength(self,commandArray,fieldName):
        return self.getRequestField(commandArray,fieldName).length
        
    def getRequestFieldOptions(self,commandArray,fieldName):
        return self.getRequestField(commandArray,fieldName).options
    
    def getRequestField(self,commandArray,fieldName):
        fields = self.getRequestFields(commandArray)
        for field in fields:
            if field.name==fieldName:
                return field
        raise CommandError(CommandError.UNKNOWN_FIELD,
                                        '%s in %s' % (fieldName, '.'.join(commandArray))) 
    
    def getRequestFields(self,commandArray):
        commandDef = self.getDefinition(self.COMMAND,commandArray)
        if 'request' not in commandDef:
            raise CommandError(CommandError.NO_REQUEST,
                                            '.'.join(commandArray)) 
        fields = [Field(fieldRaw,self.fieldOptions)
                             for fieldRaw in commandDef['request']]
        return fields
    
    def getResponseFieldNames(self,type,nameArray):
        '''
        \brief Get the response fields of a (sub)command, from its name.
        
        \param type        Command or notification?
        \param nameArray   An array of the form [commandName, subCommandname]
                           The array can be of any length, and is of length 1
                           if no subcommands are used.
        
        \exception CommandError(NO_RESPONSE) The (sub)command has
                   no request fields.
        \returns The list of request fields.
        '''
        fields = self.getResponseFields(type,nameArray)
        return [field.name for field in fields]
    
    def getResponseFieldFormat(self,type,nameArray,fieldName):
        return self.getResponseField(type,nameArray,fieldName).format
        
    def getResponseFieldLength(self,type,nameArray,fieldName):
        return self.getResponseField(type,nameArray,fieldName).length
        
    def getResponseFieldOptions(self,type,nameArray,fieldName):
        return self.getResponseField(type,nameArray,fieldName).options
    
    def getResponseField(self,type,nameArray,fieldName):
        for i in range(len(nameArray)):
            fields = self.getResponseFields(type,nameArray[:i+1])
            for field in fields:
                if field.name==fieldName:
                    return field
        raise CommandError(CommandError.UNKNOWN_FIELD,
                                        '%s in %s' % (fieldName, '.'.join(nameArray))) 
    
    def getResponseFields(self,type,nameArray):
        
        commandDef = self.getDefinition(type,nameArray)
        
        if 'response' not in commandDef:
            raise CommandError(CommandError.NO_RESPONSE,
                                            '.'.join(nameArray))
        
        keys          = list(commandDef['response'].keys())
        responseName  = keys[0]
        
        fields = [Field(fieldRaw,self.fieldOptions)
                             for fieldRaw in commandDef['response'][responseName]]
        return fields
        
    def responseFieldValueToDesc(self,nameArray,fieldName,fieldValue):
        return self.fieldValueToDesc(
                    self.COMMAND,
                    nameArray,
                    fieldName,
                    fieldValue)
            
    def notifFieldValueToDesc(self,nameArray,fieldName,fieldValue):
        return self.fieldValueToDesc(
                    self.NOTIFICATION,
                    nameArray,
                    fieldName,
                    fieldValue)
    
    def fieldValueToDesc(self,type,nameArray,fieldName,fieldValue):
        responseField = self.getResponseField(type,nameArray,fieldName)
        return responseField.options.valueToDesc(fieldValue)
    
    @classmethod
    def fieldOptionToShortDesc(self,name,value):
        for option in self.fieldOptions[name]:
            if option[0]==value:
               return option[1]
    
    @classmethod
    def fieldFormatToString(self,fieldLength,fieldFormat):
        '''
        \brief Turns the field format into a human-readable string.
        
        \param fieldLength The number of bytes (an int) of the field
        \param fieldFormat The format of a field, expressed as a string
        
        \return A human-readable string.
        '''
        returnVal = ''
        
        if   fieldFormat==FieldFormats.INT  and fieldLength==1:
            returnVal       += 'INT8U'
        elif fieldFormat==FieldFormats.INT  and fieldLength==2:
            returnVal       += 'INT16U'
        elif fieldFormat==FieldFormats.INT  and fieldLength==4:
            returnVal       += 'INT32U'
        elif fieldFormat==FieldFormats.INTS and fieldLength==1:
            returnVal       += 'INT8S'
        elif fieldFormat==FieldFormats.INTS and fieldLength==2:
            returnVal       += 'INT16'
        elif fieldFormat==FieldFormats.INTS and fieldLength==4:
            returnVal       += 'INT32'
        elif fieldFormat==FieldFormats.BOOL:
            returnVal       += 'BOOL'
        else:
            if fieldLength:
                returnVal   += "{0}B ({1})".format(fieldLength,fieldFormat)
            else:
                returnVal   += "{0}".format(fieldFormat)
        
        return returnVal
    
    #======================== validation ======================================
    
    def areSameFieldNames(self,fieldsCommand,fieldsPassed):
        '''
        \brief Determine whether the fields passed contains the same field names
               as defined in the commands.
       
        \exception CommandError.TOO_FEW_FIELDS  Too few fields are 
                   present in the fields passed.
        \exception CommandError.TOO_MANY_FIELDS Too many fields are 
                   present in the fields passed.
        \exception CommandError.UNKNOWN_FIELD At least one unknown 
                   fields in the fields passed.
        '''
        
        # list of field names expected in the command
        namesCommand    = [field[0] for field in fieldsCommand]
        namesCommand.sort()
        
        # list of field names in the passed fields
        namesPassed     = list(fieldsPassed.keys())
        namesPassed.sort()
        
        if   len(namesPassed)<len(namesCommand):
            raise CommandError(CommandError.TOO_FEW_FIELDS) 
        elif len(namesPassed)>len(namesCommand):
            raise CommandError(CommandError.TOO_MANY_FIELDS)
        else:
            for i in range(len(namesPassed)):
                if namesPassed[i]!=namesCommand[i]:
                    raise CommandError(CommandError.UNKNOWN_FIELD,namesPassed[i])
    
    def isValidFieldFormatting(self,commandArray,
                                    fieldName,
                                    fieldValue):
        '''
        \brief Determine whether the field passed contains a correct format
               according to the command definition passed.
       
        \exception CommandError.UNKNOWN_FIELD   The field is not in
                   the definition.
        \exception CommandError.MALFORMED_FIELD The field is malformed.
        '''
        thisField = self.getRequestField(commandArray,fieldName)
        
        # check whether this value is valid
        if thisField.isValidValue(fieldValue)==False:
            raise CommandError(
                CommandError.MALFORMED_FIELD,
                'commandArray={0} fieldName={1} fieldValue={2}'.format(commandArray,fieldName,fieldValue)
            )
    
    def validateRequest(self,commandArray,fields):
        '''
        \brief Validate that the fields passed form a valid request for the
               specified command or subcommand. Raises a CommandError exception
               if the request fields are invalid. 
       
        \param commandArray   An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
        \param fields         A dictionary indicating the value of every field
                              in that (sub)command, of the form
                              <tt>
                              {
                                 \<fieldName1\>: \<fieldValue1\>,
                                 \<fieldName2\>: \<fieldValue2\>,
                                  ...,
                              }
                              </tt>
       
        \exception CommandError Describes the validation error
        '''
       
        definition = self.getDefinition(self.COMMAND,commandArray)
        
        if 'request' not in definition:
            raise CommandError(CommandError.NO_REQUEST,
                                            '.'.join(commandArray))
        
        # step 1. make sure the command definition and the fields passed have
        #         the same fields (raises errors if not)
        self.areSameFieldNames(definition['request'],fields)

        # step 2. for each field, make sure formatting is correct
        #         (raises errors if not)
        for fieldName,fieldValue in fields.items():
            self.isValidFieldFormatting(commandArray,
                                        fieldName,
                                        fieldValue)
    
    #======================== serialization ===================================
    
    def _getSerializer(self,commandArray):
        '''
        \brief Get the serializer associated with a command or subcommand.
        
        \param commandArray   An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
        
        \returns The definition of a subcommand, represented as a dictionary.
        '''
        return self.getDefinition(self.COMMAND,commandArray)['serializer']
    
    def serialize(self,commandArray,fields):
        '''
        \brief Serialize a command.
       
        This function applies the serializer function specified in this
        (sub)command\'s entry in the API definition.
       
        \param commandArray   An array of the form [commandName, subCommandname]
                              The array can be of any length, and is of length 1
                              if no subcommands are used.
        \param fields         A dictionary indicating the value of every field
                              in that (sub)command, of the form
                              <tt>
                              {
                                 \<fieldName1\>: \<fieldValue1\>,
                                 \<fieldName2\>: \<fieldValue2\>,
                                  ...,
                              }
                              </tt>
       
        \returns The serialized command, in the format specified by this field\'s
                 serializer.
        '''
        
        # verify that this is a valid request (raises exception if not)
        self.validateRequest(commandArray,fields)
        
        # get the serializer associated with this (sub)command
        try:
            serialize_func = getattr(self, self._getSerializer(commandArray))
        except KeyError:
            serialize_func = self.default_serializer
            
        # apply the serializer and return the result
        return serialize_func(commandArray,fields)
    
    #======================== abstract methods ================================

    def default_serializer(self,commandArray,fields):
        '''
        \brief Default serializer for the API if no 'serializer' is specified
               in the command description
       
        The API-specific child class is expected to implement this method
        '''
        
        raise NotImplementedError('No default serializer')
    
    #======================== abstract attributes =============================
    
    ##
    # \brief Enumeration of valid options for certain fields.
    #
    # This structure is a Python dictionary. The key of each entry is an 
    # arbitrary but unique string. The value of each entry is a array of
    # options. Each option is an array of three elements:
    # - a valid value for the field
    # - a short description of that value
    # - a long description of that value
    #
    # An example redefinition is:
    # <tt>
    # fieldOptions = {
    #    'packetPriority' : [
    #       [0,     'low',                   'low'],
    #       [1,     'high',                  'high'],
    #    ],
    # }
    # </tt>
    #
    # \note While the commands variable is defined in this parent ApiDefinition
    #       class, it is meant to be redefined by inheriting classes.
    # 
    fieldOptions = {
    }
    
    ##
    # \brief Commands in this API.
    #
    # Each command is a dictionary with the following mandatory keys:
    # - 'id'           : The ID of this command (a number)
    # - 'name'         : The name of the command (a string)
    # - 'description'  : A plain-English description of the command
    #
    # The following keys are optional:
    # - 'serializer'   : a function to call to serialize this command. If this
    #                    field is absent, the default serializer will be
    #                    called.
    #
    # The remaining fields define the format of the request and response.
    #
    # In the general case, the following fields are present:
    # - 'request'      : array of fields in the request
    # - 'response'     : array of fields in the response
    #
    # A 'request' (resp. response) contains the fields contained in the request
    # (resp. response).
    #
    # \note Commands can define only a request (resp. response), in which case
    #       the response (resp. request) field is not present.
    #
    # In some cases, a command does not defines request/response directly, but 
    # rather subcommands. In that case, the following fields are present:
    # - 'subcommands'  : The name of the dictionary subcomands for this
    #                    function
    #
    # \note While the commands variable is defined in the parent API class,
    #       it is meant to be redefined by classes inheriting from this one.
    #
    commands = [
    ]
    
    ##
    # \brief Notifications in this API.
    #
    # \note While the notifications variable is defined in the parent API class,
    #       it is meant to be redefined by classes inheriting from this one.
    #
    notifications = [
    ]



"""
This module defines the _RemoteGatewayClient class used as a singleton
in the streamlit application process.

It will be used to recevie/send inter process messages from/to the Gateway
process.
"""
import threading
import time
from multiprocessing.connection import Connection
import streamlit as st
from streamlit.scriptrunner.script_run_context import add_script_run_ctx as add_report_ctx
from streamlit.server.server import Server
from txp.common.config import settings
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


# ==============================================================================
# Definition of the _RemoteGatewayClient.
# ==============================================================================
class _RemoteGatewayClient:
    """This Remote Gateway client encapsulates the necessary multiprocessing utilities
    to communicate with the Gateway main process in streamlit app.

    This object is designed to live in the streamlit application
    process, as a singeton.
    """

    def __init__(self):
        """
        Args:
            gateway_connection: The multiprocessing.Connection object to send/receive IPC objects
                to/from the gateway process.
        """
        self._gateway_connection: Connection = None

        self._connection_update_thread: threading.Thread = None
        """_connection_update_thread: A thread designed to run in background, 
        receiving the newest state from the Gateway."""

        self._finish_update_event: threading.Event = threading.Event()

        self.last_known_state = None
        """last_known_state: This helps to launch the streamlit application, only
        when a state was received from the Gateway."""

    def set_gateway_connection(self, gateway_connection: Connection) -> None:
        """Sets the Gateway connection in the UI process."""
        log.debug(
            f"_RemoteGatewayClient receiving Gateway Connection: "
            f"{id(gateway_connection)}"
        )
        self._gateway_connection = gateway_connection

    def _connection_update_loop(self, pill: threading.Event):
        """A loop to receive states from the Gateway, and refresh the streamlit
        application programmatically.

        Note: Using a workaround to refresh the view. Streamlit doesn't support yet
        a public stable method to re-run the script on command. There's an experimental
        feature: experimental_rerun. But it didn't seem to work outside the main thread.

        Todo: If streamlit moves the experimental_rerun to the stable API, revisit this
            implementation.
        """
        while True:
            while not self._gateway_connection.poll():
                time.sleep(0.5)  # waits 0.1 seconds to ask again if there's a new state
                if pill.is_set():
                    log.debug("Closing update thread")
                    return

            try:
                new_state_from_gateway = (
                    self._gateway_connection.recv()
                )  # blocks until there's something new to receive

            except:
                # The recommended way is for only one client thread to publish states to the streamlit process.
                # Just in case, to avoid runtime problems, catching an exception.
                # TODO: We should implement a recovery strategy of the pipes in the future
                log.error('Error receiving UI in streamlit process. The UI will not be updated')
                continue

            self.last_known_state = new_state_from_gateway
            st.session_state.gateway_state = new_state_from_gateway
            log.debug(f"New Gateway state received: {new_state_from_gateway}")

            # st.experimental_rerun doesn't seems to work properly with threads.
            #try:
            #    st.experimental_rerun()
            #except st.script_runner.RerunException as e:
            #    log.warning("Exception trying to rerun streamlit app.")

            # This refresh snippet is taken from:
            # https://github.com/streamlit/streamlit/issues/3854

            # Streamlit <= 1.3.0 should use this commented code to get the session ID.
            # ctx = st.report_thread.get_report_ctx()
            # session_id = add_report_ctx().streamlit_script_run_ctx.session_id
            current_server = Server.get_current()
            session_id = add_report_ctx().streamlit_script_run_ctx.session_id
            session = current_server.get_session_by_id(session_id)
            log.debug(f"Streamlit session ID: {session_id}")
            if session is None:
                # If the session is None, then a new tab was opened and that's the new
                # session to be updated in real time.
                log.debug("The streamlit session was not found.")
            else:
                session.request_rerun(None)

    def start_refresh_loop(self) -> None:
        """Starts a receive background thread responsible of receiving new states from
        the Gateway process.
        """
        # First, it'll end the running thread for the previous running session.
        # That session is assumed to be closed. It will not receive new real time updates.
        if self._connection_update_thread is not None:
            self._finish_update_event.set()
            while self._connection_update_thread.is_alive():
                pass
            self._connection_update_thread = None
            self._finish_update_event = threading.Event()

        self._connection_update_thread = threading.Thread(
            target=self._connection_update_loop,
            name="_gateway_connection_update_thread",
            daemon=True,
            args=(self._finish_update_event,)
        )
        add_report_ctx(
            self._connection_update_thread
        )  # Add the streamlit context report to the thread
        self._connection_update_thread.start()

    def wait_for_initial_state(self) -> None:
        """Waits to receive the initial state from the Gateway.

        It should be called before starting the streamlit app to
        have a first state to render.
        """
        new_state_from_gateway = (
            self._gateway_connection.recv()
        )  # blocks until there's something new to receive
        self.last_known_state = new_state_from_gateway

    def send_state_to_gateway(self, ux_state) -> None:
        """Sends the ux_state message to the Gateway."""
        log.debug(f"Send UxState to Gateway: {ux_state}")
        self._gateway_connection.send(ux_state)


# ==============================================================================
# Declaration of the _RemoteGatewayClient singleton instance.
# ==============================================================================

remote_gateway_client: _RemoteGatewayClient = _RemoteGatewayClient()



"""
This module implements the Device Manager functionality.
The main functionality of the Device Manager is to centralize the
lifecycle of DriversHost processes and to provide collaboration
with higher components like Gateway and user interaction components.
"""


# ============================ imports =========================================
import multiprocessing
import time
from typing import List, Dict, Tuple, Optional

from txp.common.edge import EdgeDescriptor
from txp.devices.drivers.drivers_host import DriversHost
from txp.devices.drivers.icomox.icomox_driver_host import IcomoxDriversHost
from txp.devices.drivers.mock.mock_driver import MockDriversHost
from txp.devices.drivers.voyager.voyager_driver_host import VoyagerDriversHost
from txp.devices.drivers.usb.usb_driver_host import UsbDriversHost
from txp.devices.drivers.generic_driver_host import GenericDriversHost
from txp.devices.drivers.arm_robot.robot_host import RobotDriversHost, ThermalArmHost
from txp.devices.package_queue import PackageQueueProxy
from txp.devices.drivers.device_type import DeviceType
from multiprocessing import managers
from txp.common.config import settings
import itertools
import logging
from txp.devices.sampling_task import SamplingTask
import gc

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class _HostsProcessManager(managers.BaseManager):
    """Internal class used as custom multiprocessing.Manager to create server processes
    for the DriversHost's instances."""

    _host_registered: bool = False

    @classmethod
    def register_hosts_classes(cls):
        """Registers the types and callables for the different DriversHost implementations
        in this multiprocessing Manager class.
        """
        if not cls._host_registered:
            log.debug("Registering DriversHosts in the HostProcessManager")
            cls.register("VoyagerDriversHost", VoyagerDriversHost)
            cls.register("IcomoxDriversHost",IcomoxDriversHost)
            cls.register("MockDriversHost", MockDriversHost)
            cls.register("UsbDriversHost", UsbDriversHost)
            cls.register("GenericDriversHost", GenericDriversHost)
            cls.register("RobotDriversHost", RobotDriversHost)
            cls.register("ThermalArmHost", ThermalArmHost)
            cls._host_registered = True


# ==============================================================================
# DeviceManager class to be used at initialization time.
# ==============================================================================
class DevicesManager:
    """This class is the top-level manager used to manage the lifecycle of the
    DriversHost processes for the devices connected to the Gateway.

    This DeviceManager provides collaboration methods that are used by higher level
    components such as:
        - Gateway component
        - User interaction component for pairing logic
    """

    def __init__(self):
        _HostsProcessManager.register_hosts_classes()

        self._hosts_processes: List[_HostsProcessManager] = []
        """_hosts_processes: A List of the used _HostsProcessManagers created to 
        serve the DriversHosts."""
        self.hosts: List[DriversHost] = []
        """hosts: the set of DriversHost references created by this device manager."""
        self.drivers_status_queue = multiprocessing.Manager().Queue()

        self._driver_to_host: Dict[str, DriversHost] = {}
        """_driver_to_host: internal map to known in which host reference lives a Driver."""

        self._virtual_devices_table: Dict[str, bool] = {}
        """_virtual_devices_table: this dict is used internally to know which logical id
            is virtual or not."""

        self._drivers_clock_values: Dict[str, int] = {}
        """_devices_clock_values: this dict is used internally to know which is the value
        of a driver instance clock.
        """

    def create_host(
        self,
        host_device: str,
        gateway_package_queue: PackageQueueProxy,
        host_kwargs: Dict = {},
    ):
        """Create the instance for the specified DriversHosts.

        This allows client components to decide which devices hosts needs to be created
        based on their own business logic.

        Args:
            host_device: name of the device equals to a Driver.device_name() implementation.
            gateway_package_queue: The Gateway central package Queue to be distributed
                to all the DriersHosts.
            host_kwargs: Dict of specific parameters for the DriversHost device implementation.

        Returns:
            None
        """
        process_manager = _HostsProcessManager()
        process_manager.start()
        process_manager.connect()
        host = DeviceType(host_device).build_host_process(process_manager, gateway_package_queue, self.drivers_status_queue, **host_kwargs)
        self.hosts.append(host)
        self._hosts_processes.append(process_manager)

    def initialize_drivers(self, edges: List[EdgeDescriptor]):
        """
        This funtion initializes all edges configured.
        """
        for host in self.hosts:
            edges_filtered = list(filter(lambda edge: self._get_drivers_by_host(host.get_driver_host_name(), edge), edges))
            host.add_drivers(edges_filtered)
            self._driver_to_host.update({
                edge.logical_id: host for edge in edges_filtered
            })
            self._virtual_devices_table.update({
                edge.logical_id: self.is_virtual_driver(edge.logical_id)
                for edge in edges_filtered
            })
            self._drivers_clock_values.update({
                edge.logical_id: host.get_driver_observation_pulse(edge.logical_id)
                for edge in edges_filtered
            })

    def start(self):
        """
        Starts all registered edges related connection handlers. DriverHost associated registered instances.

        """
        try:
            for host in self.hosts:
                host.connect()
        except Exception as ex:
            log.error("Error starting gateway host: {}".format(host.get_driver_host_name()))
            raise

    def stop(self):
        """
        Stops all registered edges related connection handlers. DriverHost associated registered instances.

        """
        for host in self.hosts:
            host.disconnect()

    def clear_devices(self):
        """Clear the DriversHost from the device manager.

        The hosts will be stopped first. After that, the manager processes
        are explicitly "shutdown". Finally, the references are cleared
        from the DeviceManager, in order to be garbage collected.
        """
        if self.hosts:
            log.info("Stopping DeviceManager hosts")
            self.stop()

            log.info("Clearing DeviceManager hosts")
            self.hosts = []

            for manager in self._hosts_processes:
                manager.shutdown()

            self._hosts_processes = []
            self._driver_to_host.clear()
            self._virtual_devices_table.clear()
            self._drivers_clock_values.clear()

            gc.collect()

        else:
            log.info("DevicesManager does not have hosts to clear.")

    def get_state_changed_queued_items(self):
        """
        gets that contains serial and state of the edge that changed its state.
        :return: a list of StateReports. Empty list if no queue state changed
        """
        edges_states_reports = []
        while self.drivers_status_queue.qsize():
            state_report = self.drivers_status_queue.get(False)
            edges_states_reports.append(state_report)
            log.debug(edges_states_reports)

        return edges_states_reports

    def receive_sampling_tasks(self, tasks: Dict[str, SamplingTask]):
        for host in self.hosts:
            host.receive_sampling_tasks(tasks)

    def receive_sampling_tasks_list(self, tasks: List[Tuple]) -> None:
        """Receive the sampling tasks lists and send those tasks to the drivers."""
        tasks_with_host = []
        for logical_id, task in tasks:
            tasks_with_host.append((logical_id, task, self._driver_to_host[logical_id]))

        for host, entries in itertools.groupby(tasks_with_host, lambda entry: entry[2]):
            log.debug(f"Host {host.get_driver_host_name()} will receive the tasks: {entries}")
            tasks_for_host = list(map(
                lambda entry: (entry[0], entry[1]),
                entries
            ))
            host.receive_sampling_tasks_list(tasks_for_host)

    def notify_sampling_window_completed(self, logical_ids: Optional[List[str]] = None):
        """Notifies to the drivers that the current sampling window was completed.

        Args:
            logical_ids: An optional list of drivers to enable. If not provided,
                then all drivers in the device manager will be enabled.
        """
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to notify sampling window completion"
                          f"the device: {logical_id}, but the device was not found")
            else:
                self._driver_to_host[logical_id].notify_sampling_window_completed(logical_id)

    def start_sampling_windows(self, logical_ids: Optional[List[str]] = None):
        """Starts the drivers internal sampling windows loop.

        Args:
            logical_ids: An optional list of drivers to enable. If not provided,
                then all drivers in the device manager will be enabled.
        """
        if not logical_ids:
            for host in self.hosts:
                host.start_sampling_windows()
        else:
            for logical_id in logical_ids:
                if logical_id not in self._driver_to_host:
                    log.error("Devices Manager requested to enable a windows loop for "
                              f"the device: {logical_id}, but the device was not found")
                else:
                    self._driver_to_host[logical_id].start_sampling_windows(logical_id)

    def stop_sampling_windows(self, logical_ids: Optional[List[str]] = None):
        """Stops the drivers intenral sampling windows loop.

        Args:
            logical_ids: An optional list of drivers to disable. If not provided,
                then all drivers in the device manager will be call to disable.
        """
        if not logical_ids:
            for host in self.hosts:
                host.stop_sampling_windows()
        else:
            for logical_id in logical_ids:
                if logical_id not in self._driver_to_host:
                    log.error("Devices Manager requested to disable a windows loop for "
                              f"the device: {logical_id}, but the device was not found")
                else:
                    self._driver_to_host[logical_id].stop_sampling_windows()

    def reset(self,host_device):
        """
        Intended to reset an edge type.

        :param host_device: Driver.get_name attribute to reset this type of edge

        """
        for host in self.hosts:
            if host_device == host.get_driver_host_name():
                host.disconnect()
                time.sleep(5)
                host.connect()

    def _get_drivers_by_host(self, driver_name: str, edge: EdgeDescriptor):
        edge_descriptor = edge
        if driver_name == edge_descriptor.device_kind:
            return True
        else:
            return False

    def get_drivers_observation_pulse(self, logical_ids: List[str]) -> List:
        """Returns the observation pulses of the corresponding drivers."""
        ret = []
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to obtain pulse from a "
                          "Host that has not been created")
                continue

            pulse = self._driver_to_host[logical_id].get_driver_observation_pulse(logical_id)
            ret.append((logical_id, pulse))
        return ret

    def get_driver_observation_pulse(self, logical_id: str) -> int:
        """Returns the observation pulses of the driver."""
        if logical_id not in self._drivers_clock_values:
            log.error("Devices Manager requested to obtain clock value from a "
                      f"unknown driver {logical_id}")
            return 0

        return self._drivers_clock_values[logical_id]

    def enable_drivers_clock_pulse(self, logical_ids: List[str]) -> None:
        """Enables the distributed drivers to execute an observation pulse."""
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to obtain pulse from a "
                          "Host that has not been created")
                continue

            self._driver_to_host[logical_id].enable_driver_clock_pulse(logical_id)

    def disable_drivers_clock_pulse(self, logical_ids: List[str]) -> None:
        """Disables the drivers current clock pulse sample collect"""
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to obtain pulse from a "
                          "Host that has not been created")
                continue

            self._driver_to_host[logical_id].disable_current_pulse(logical_id)

    def get_drivers_completed_clock(self, logical_ids: List[str]) -> Dict[str, int]:
        d = {}
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to obtain completed clocks from a "
                          "Host that has not been created")
            else:
                p = self._driver_to_host[logical_id].get_driver_completed_pulses(logical_id)
                log.debug(f"Host reported {p} completed clock pulses for {logical_id}")
                d[logical_id] = p

        return d

    def clear_drivers_completed_clocks(self, logical_ids: List[str]) -> None:
        """Clear the drivers completed pulses registered in the drivers hosts.
        """
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to clear completed clocks register from a "
                          "Host that has not been created")
            else:
                self._driver_to_host[logical_id].clear_driver_registered_packages(logical_id)
                log.debug(f"Clearing completed registered packages for driver {logical_id}")

    def is_virtual_driver(self, logical_id: str) -> bool:
        if logical_id not in self._driver_to_host:
            log.error("Devices Manager requested to obtain virtual type from a "
                      "Host that has not been created")
            return False
        return self._driver_to_host[logical_id].is_virtual_driver(logical_id)

    def is_actuator_device(self, logical_id: str) -> bool:
        """Returns true if the logical_id provided belongs to an
        actuator device type.

        The only supported expected actuator device is a logical id
        for a Robotic arm.
        """
        if logical_id not in self._driver_to_host:
            log.error(f"Devices Manager was requested to check driver actuator type "
                      f"from driver {logical_id}, but the Driver was not found")
            return False
        return self._driver_to_host[logical_id].is_actuator_device(logical_id)

    def get_drivers_completed_clock(self, logical_ids: List[str]) -> Dict[str, int]:
        d = {}
        for logical_id in logical_ids:
            if logical_id not in self._driver_to_host:
                log.error("Devices Manager requested to obtain completed clocks from a "
                          "Host that has not been created")
            else:
                p = self._driver_to_host[logical_id].get_driver_completed_pulses(logical_id)
                log.debug(f"Host reported {p} completed clock pulses for {logical_id}")
                d[logical_id] = p

        return d



from txp.common.ml.models import ModelStateValue, ModelRegistry
from txp.ml.training_service.common import TrainCommand
from txp.web.core_components.main_view import MainView
from txp.web.core_components.app_profiles import AppProfile
from fastapi import status
import streamlit as st
import logging
from txp import cloud as cloud_utils
import requests
from txp.cloud.common.config import settings
from txp.cloud.controllers.firestore import *

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


class TrainingView(MainView):
    """This view is used for training models based on datasets available
    in the data warehouse."""

    def __init__(self):
        super(TrainingView, self).__init__(component_key="training_view")
        self._asset_selectbox_val: str = ""
        self._asset_selectbox_key: str = "asset_sb"
        self._task_selectbox_val: str = ""
        self._task_selectbox_key: str = "task_sb"
        self._dataset_name_val: str = ""
        self._dataset_name_key: str = "dataset_name_key"
        self._versions_val: List[str] = []
        self._versions_key: str = "versions_form_key"
        self._project_configuration_data: ProjectModel = None
        self._model_registry_col: str = "txp_models_registry"
        self._pulled_model: ModelRegistry = None

    def get_associated_profile(self) -> AppProfile:
        return AppProfile.ANALYTICS

    @classmethod
    def get_display_name(cls) -> str:
        return "Entrenar Task"

    def _build_submenu_state(self) -> Dict:
        return {}

    def _build_content_state(self) -> Dict:
        return {}

    def _get_project_configuration(self):
        db = get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )
        self._project_configuration_data = get_current_project_model(
            db, self._tenant_id
        )
        self._current_gateway = list(
            self._project_configuration_data.gateways_table.keys()
        )[0]

    def _register_asset_value(self):
        self._asset_selectbox_val = st.session_state[self._asset_selectbox_key]

    def _register_dataset_name_value(self):
        self._dataset_name_val = st.session_state[self._dataset_name_key]

    def _register_dataset_versions_values(self):
        self._versions_val = (
            st.session_state[self._versions_key].replace(" ", "").split(",")
        )

    def _register_task_value(self):
        self._task_selectbox_val = st.session_state[self._task_selectbox_key]

    def _validate_training_allowed(
        self, dataset_name: str, dataset_versions: List[str]
    ) -> bool:
        firestore_db: firestore.Client = get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )
        model_registry_name = cloud_utils.get_gcs_dataset_prefix(
            dataset_name, dataset_versions
        )
        model = cloud_utils.get_ml_model(
            firestore_db,
            self._tenant_id,
            self._asset_selectbox_val,
            self._task_selectbox_val,
            model_registry_name,
            self._model_registry_col,
        )
        firestore_db.close()
        if not model:
            return True

        if model.state.value.value == ModelStateValue.TRAINING.value:
            return False

        else:
            return True

    def _train_model(self, train_payload: TrainCommand):
        headers = {"Authorization": f"Bearer {st.secrets['service_auth_token']}"}
        response = requests.post(
            f"{st.secrets.service_url}/training/train", json=train_payload.dict(),
            headers=headers
        )
        if response.status_code == status.HTTP_200_OK:
            st.success("Su modelo se ha empezado a entrenar correctamente")

        elif response.status_code == status.HTTP_404_NOT_FOUND:
            st.warning(f"Error al lanzar entrenamiento: {response.json()['Error']}")
        else:
            st.warning(f"Error desconocido al lanzar entrenamiento: ")
            st.json(response.json())

    def _publish_model(self, publish_payload: TrainCommand):
        headers = {"Authorization": f"Bearer {st.secrets['service_auth_token']}"}
        response = requests.post(
            f"{st.secrets.service_url}/training/publish", json=publish_payload.dict(),
            headers=headers
        )
        if response.status_code == status.HTTP_200_OK:
            st.success("Su modelo ha sido promovido correctamente al servicio de predicción")
            self._pulled_model = None  # Clear the memory from memory

        elif response.status_code == status.HTTP_404_NOT_FOUND:
            st.warning("Su modelo no existe o ya esta publicado en el servicio de predicción")
            st.text("Si el problema persiste, contacte a soporte.")

    def _render_content(self) -> None:
        if not self._project_configuration_data:
            self._get_project_configuration()

        st.markdown("Entrenar Task para un Activo")
        logging.info(f"Rendering selected active: {self._asset_selectbox_val}")
        logging.info(f"Rendering selected task: {self._task_selectbox_val}")

        self._asset_selectbox_val = st.selectbox(
            label="Seleccione Activo",
            options=list(self._project_configuration_data.machines_table.keys()),
            key=self._asset_selectbox_key,
            on_change=self._register_asset_value,
            index=0,
        )

        self._task_selectbox_val = st.selectbox(
            label="Seleccione Task",
            options=list(
                self._project_configuration_data.tasks_by_asset[
                    self._asset_selectbox_val
                ].keys()
            ),
            on_change=self._register_task_value,
            key=self._task_selectbox_key,
            index=0,
        )

        with st.form(key="dataset_info_form"):
            dataset_name = st.text_input(
                label="Ingrese el nombre del dataset",
                help="El nombre del dataset acordado por el equipo de anotación",
                key=self._dataset_name_key,
            )

            dataset_versions = st.text_input(
                label="Ingrese las versiones separadas por coma", key=self._versions_key
            )

            st.form_submit_button("Entrenar", on_click=self._start_train)

            st.form_submit_button(
                "Ver resultado",
                on_click=self._check_train_result,
            )

        if self._pulled_model:
            st.markdown("**Resumen del modelo:**")
            st.markdown(f"Tenant: {self._pulled_model.metadata.tenant_id}")
            st.markdown(f"Task: {self._pulled_model.metadata.task_id}")
            st.markdown(f"Asset: {self._pulled_model.metadata.machine_id}")
            st.markdown(f"Estado del modelo:")
            st.json(self._pulled_model.state.json())
            st.markdown(f"Feedback del modelo:")
            st.json(self._pulled_model.metadata.feedback.json())

            if self._pulled_model.state.value == ModelStateValue.ACKNOWLEDGE:
                st.info(
                    "El modelo ha terminado de entrenar, y puede ser promovido a predicción"
                )
                st.button("Promover", on_click=self._approve_ack_model)

    def _start_train(self):
        st.info("Starting training process...")
        self._register_asset_value()
        self._register_task_value()
        self._register_dataset_versions_values()
        self._register_dataset_name_value()

        training_allowed = self._validate_training_allowed(
            self._dataset_name_val, self._versions_val
        )
        if not training_allowed:
            st.warning(
                "Actualmente su Task se encuentra entrenando. "
                "Espere a que el task termine de entrenar"
            )
        else:
            train_payload = TrainCommand(
                dataset_name=self._dataset_name_val,
                dataset_versions=self._versions_val,
                tenant_id=self._tenant_id,
                machine_id=self._asset_selectbox_val,
                task_id=self._task_selectbox_val,
            )
            self._train_model(train_payload)
            self._pulled_model = None

    def _check_train_result(self):
        self._register_asset_value()
        self._register_task_value()
        self._register_dataset_versions_values()
        self._register_dataset_name_value()
        firestore_db: firestore.Client = get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )
        model_reg_name = cloud_utils.get_gcs_dataset_prefix(
            self._dataset_name_val, self._versions_val
        )
        logging.info(
            "Requesting model from models registry for: \n"
            f"Tenant: {self._tenant_id} \n"
            f"Asset: {self._asset_selectbox_val} \n"
            f"Task: {self._task_selectbox_val} \n"
            f"Model Registry: {model_reg_name}"
        )
        model = cloud_utils.get_ml_model(
            firestore_db,
            self._tenant_id,
            self._asset_selectbox_val,
            self._task_selectbox_val,
            model_reg_name,
            self._model_registry_col,
        )
        if model is None:
            st.error(
                "No se pudo encontrar registros para el modelo. Si el problema persiste, "
                "contacte a soporte"
            )
            self._pulled_model = None
            return

        else:
            self._pulled_model = model

    def _approve_ack_model(self):
        train_payload = TrainCommand(
            dataset_name=self._dataset_name_val,
            dataset_versions=self._versions_val,
            tenant_id=self._tenant_id,
            machine_id=self._asset_selectbox_val,
            task_id=self._task_selectbox_val,
        )
        self._publish_model(train_payload)

    def _render_submenu(self) -> None:
        if not self._project_configuration_data:
            self._get_project_configuration()



"""
The streamlit entrypoint for the streamlit cloud application.
Currently this is a "Hello World" used to test the txp classes in the
streamlit cloud deployment.
"""
# ============================ imports =========================================
from txp.web.core_components.main_app_component import MainAppComponent
from txp.web.views import *

import streamlit as st

st.set_page_config(layout="wide")


# ========================= Main App ============================================
def initialize_app():
    if 'initialized' not in st.session_state:
        app_component = MainAppComponent()  # Create the layout component
        app_component.add_child_component(AssetConditionsView(), make_current=True)
        app_component.add_child_component(DeviceHealthView())
        # app_component.add_child_component(TelemetryHealthView())
        app_component.add_child_component(AnalyticsView())

        st.session_state.initialized = True
        st.session_state.app_layout = app_component


initialize_app()
st.session_state.app_layout.render()



"""
This Script can be used to create custom claims on some registered user in Identity Platform.

For the script to run, remember to add the environment variable
for the Default App engine service account JSON.

Instructions:
    1. Call the script with the following parameters in this order:
        - MAIN email of the user to set the claims
        - claim to set on the user

The email must be the main email from the user in Identity Platform:
    https://firebase.google.com/docs/auth/admin/manage-users#retrieve_user_data

Example with one claim:
    python script_create_custom_claims.py Admin

"""
import sys
import firebase_admin
from firebase_admin import auth as firebase_auth
from firebase_admin.exceptions import FirebaseError


if __name__ == '__main__':
    default_app = firebase_admin.initialize_app()
    user_email = sys.argv[1]
    print(f"Trying to create claims for the user with email: {user_email}")

    try:
        user = firebase_auth.get_user_by_email(user_email)
    except ValueError as e:
        print(f'There was an error, the email is invalid or malformed: {e}')
        exit(1)
    except firebase_auth.UserNotFoundError:
        print(f'The user was not found.')
        exit(1)
    except FirebaseError as e:
        print(f'There was an unknown error: {e}')
        exit(1)

    print(f"User found: {user.display_name}")

    claims = {'role': sys.argv[2]}

    print(f"The following claims will be configured: {claims}")

    try:
        firebase_auth.set_custom_user_claims(user.uid, claims)

    except ValueError:
        print("The claims are invalid.")
        exit(1)
    except FirebaseError as e:
        print(f'There was an unknown error: {e}')
        exit(1)

    print(f"SUCCESS: The claims were updated for the user.")

    # user_email = sys.argv[1]
    # user = firebase_auth.create_user(
    #     email=user_email,
    #     email_verified=True
    # )



"""
This module implements the Generic common AppComponent interface
for all the components shown in the application:
    - Main App component
    - Sidebar Component (left)
    - Main View Component
"""
import abc
import logging
from typing import Dict, List, Union

from txp.cloud.controllers.firestore import ProjectModel
from txp.web.core_components.app_profiles import AppProfile
from txp.cloud.controllers.authentication import AuthenticationInfo
import txp.cloud.controllers.firestore as firestore_controller
from txp.web.core_components.states_and_events_report import *
from collections import defaultdict
from txp.web.resources.resources_management import ImageManagement, VideoManagement
import pandas as pd
import markdown

# TODO: This might be a problem. The logging configuration is being taken from `txp` package
from txp.common.config import settings

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


@dataclasses.dataclass
class AppState:
    current_profile: AppProfile = None
    current_view: str = ""
    views_for_profile: Dict[str, List[str]] = dataclasses.field(default_factory=dict)
    authentication_info: AuthenticationInfo = dataclasses.field(
        default_factory=AuthenticationInfo
    )
    condition_report: ConditionReport = dataclasses.field(
        default_factory=ConditionReport
    )
    event_report: EventReport = dataclasses.field(default_factory=EventReport)
    project_model: ProjectModel = None

    def __post_init__(self):
        self.views_for_profile = defaultdict(list)


@dataclasses.dataclass
class GridCell:
    """A GridCell currently can hold:
        - Markdown

    In theory, anything that can be injected in HTML can be shown in the grid.
    """

    css_class: str
    grid_column_start: int
    grid_column_end: int
    grid_row_start: int
    grid_row_end: int
    width: str = "auto"

    def __post_init__(self):
        self.inner_html = ""

    def _to_style(self) -> str:
        """Returns the grid related css style for this class"""
        return f"""
        .{self.css_class} {{
            grid-column-start: {self.grid_column_start};
            grid-column-end: {self.grid_column_end};
            grid-row-start: {self.grid_row_start};
            grid-row-end: {self.grid_row_end};
        }}
        """

    def text(self, text: str = ""):
        """Renders text in HTML as plain text"""
        self.inner_html = text

    def markdown(self, text: str = ""):
        self.inner_html = markdown.markdown(
            text, extensions=["tables"]
        )  # tables support

    def image(self, source, alt="Image", width=100, height=100):
        """
        Render an HTML img object.
        """
        self.inner_html = (
            f"""<div><img src={source} width={width} height={height} alt={alt}></div>"""
        )

    def _to_html(self):
        return f"""<div class="box {self.css_class}">{self.inner_html}</div>"""


@dataclasses.dataclass
class Grid:
    """A class to handle a CSS Grid

    The idea for this implementation was taken from here:
        https://github.com/MarcSkovMadsen/awesome-streamlit/blob/be454e29c35a9a1a760b1737a5176a47f4f9717b/gallery/layout_experiments/app.py#L79

    This grid can render not only markdown, but other types of web content.
    """

    template_column: str
    gap_between_columns: int
    background_color: str
    color: str
    font_size: str
    tables_font_size: str
    cells: List

    def __enter__(self):
        """__enter__ method to allow `with` statement"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """After a `with` statement, writes the styles to the html"""
        st.markdown(self._get_grid_style(), unsafe_allow_html=True)
        st.markdown(self._get_cells_style(), unsafe_allow_html=True)
        st.markdown(self._get_cells_html(), unsafe_allow_html=True)

    def _get_grid_style(self):
        return f"""
        <style>
            .wrapper {{
            display: grid;
            grid-template-columns: {self.template_column};
            grid-gap: {self.gap_between_columns};
            background-color: white;
            color: {self.color};
            width: auto;
            }}
            .box {{
            min-height: 20px
            background-color: white;
            color: black;
            border-radius: 0px;
            padding: 10px;
            margin-top: 0px;
            margin-left: 5px;
            margin-right: 5px;
            margin-bottom: 0px,
            font-size: {self.font_size};
            }}
            table {{
                color: {self.color};
                font-size: {self.tables_font_size};
                width: 100%;
            }}
            /* content-table css class is inserted by DataFrame.to_html() in Card class */
            .content-table thead tr {{
            background-color: #313131;
            color: #ffffff;
            font-weight: bold;
            }}
            .content-table th,
            .content-table td {{
            text-align: center;
            }}
            .content-table tbody tr {{
            border-bottom: 1px solid #d1c9ca;
            }}
            .centrado {{
            width:100%;
            max-width:600px;
            }}
            
        </style>
        """

    def _get_cells_style(self):
        return (
            "<style>"
            + "\n".join([cell._to_style() for cell in self.cells])
            + "</style>"
        )

    def _get_cells_html(self):
        return (
            '<div class="wrapper">'
            + "\n".join([cell._to_html() for cell in self.cells])
            + "</div>"
        )

    def add_cell(
        self,
        css_class: str,
        grid_column_start,
        grid_column_end,
        grid_row_start,
        grid_row_end,
    ):
        cell = GridCell(
            css_class=css_class,
            grid_column_start=grid_column_start,
            grid_column_end=grid_column_end,
            grid_row_start=grid_row_start,
            grid_row_end=grid_row_end,
        )
        self.cells.append(cell)
        return cell


@dataclasses.dataclass
class Card:
    """A class for standardizing the view composition.
    Parameters:
        card_row(int): row number of the view where the card will be painted.
        card_column_end(int): column number of the view where the card will end.
        card_subheader(str): card subheader.
        labels(list): list of labels for the content table.
        content(dict): dict of items for the content table.
        vertical_style(boolean): if the value is True, the table will have
                the labels in the index.
    """

    def __init__(
        self,
        card_row: int,
        card_column_start: int,
        card_column_end: int,
        card_subheader: str,
        content: Union[str, dict],
        labels=[],
        vertical_style=False,
    ):
        self.row = card_row
        self.column_start = card_column_start
        self.column_end = card_column_end
        self.card_subheader = card_subheader
        self.content = content

        if type(self.content) == dict:
            self.table = self._get_table_component(labels, self.content, vertical_style)

    def _get_table_component(
        self,
        labels,
        content,
        vertical_style,
    ):
        """Method to build the table of contents with the labels and items.
        Parameters:
                labels(list): list of column names for the content table.
                content(dict): dict of items for de content table.
                vertical_style(boolean): if the value is True, the table will have
                the labels in the index.
        Returns:
            content table for card.
        """
        df = pd.DataFrame(content)
        df.columns = labels

        if vertical_style:
            df_t = df.transpose()
            df_t.columns = [""] * len(df_t.columns)
            table = df_t.to_html(
                classes="content-table",
                escape=False,
            )
            return f"""
{table}"""
        else:
            table = df.to_html(classes="content-table", escape=False, index=False)
            return f"""
{table}"""

    def _build_card(self):
        """Method to build a card.
        Returns:
              f-string with components of card."""

        if type(self.content) == dict:
            content = f"""{self.table}"""

        else:
            content = self.content

        subheader_card = f"""##### {self.card_subheader}  \n"""
        card = subheader_card + content
        return card


class AppComponent(abc.ABC):
    def __init__(self, component_key: str):
        """
        Args:
            component_key(str): The component key to be used in logging and as
                key in the streamlit state if necessary.
        """
        self._component_key = component_key
        self._components: Dict[str, AppComponent] = {}
        st.session_state[
            self._component_key
        ] = self  # adds the component to streamlit state

        # Load from the secrets all the useful parameters for the components internal workings.
        self._tenant_id: str = st.secrets["tenant_id"]
        self._bigquery_dataset: str = st.secrets['big_query_dataset']
        self._events_dataset: str = st.secrets["events_dataset"]
        self._states_dataset: str = st.secrets["states_dataset"]
        self._days_report: int = st.secrets["days_report"]
        self._edge_states_sp = {
            "Disconnected": "Desconectado",
            "Connected": "Conectado",
            "Sampling": "Recolectando datos",
        }
        self._asset_states_sp = {
            "OPTIMAL": "Óptima",
            "GOOD": "Buena",
            "UNDEFINED": "Indefinida",
            "OPERATIONAL": "Operativa",
            "CRITICAL": "Crítica",
        }

    # App Global State handling methods
    @property
    def app_state(self) -> AppState:
        """Gets the global state of the application"""
        return st.session_state["app_state"]

    @classmethod
    def _register_app_state(cls, state: AppState) -> None:
        st.session_state["app_state"] = state

    def render(self) -> None:
        """Renders this component next in the streamlit app run."""
        log.debug(f"Calling render of component {self._component_key}")
        self._render()

    @abc.abstractmethod
    def _render(self) -> None:
        """This method implementation defines the visual drawing of the component
        in the browser, in a sequential way.

        When this method is called, streamlit will render according to the instructions
        declared in the method.
        """
        pass

    def render_grid(
        self,
        template_column: str = "1 1 1",
        gap_between_columns="5px",
        background_color="rgb(255, 255, 255)",
        color="black",
        font_size="125%",
        tables_font_size="15px",
    ) -> Grid:
        """Returns a Grid to render a grid of cells."""
        return Grid(
            template_column,
            gap_between_columns,
            background_color,
            color,
            font_size,
            tables_font_size,
            [],
        )

    def render_card(self, cards: list):
        """Method to paint one or more cards in a row view.
        Parameters:
                cards (list): list of cards.

        Returns:
                row of view with one or more cards."""

        with self.render_grid() as grid:
            for index, card in enumerate(cards):
                grid.add_cell(
                    css_class=f"card_{index}",
                    grid_column_start=card.column_start,
                    grid_column_end=card.column_end,
                    grid_row_start=card.row,
                    grid_row_end=card.row,
                ).markdown(f"""{card._build_card()}"""),

        st.markdown("***")

    def image_html(self, file_name, responsive=False, width=60):
        """
        returns an image html tag from a file_name
        """
        if responsive:
            resources = ImageManagement(file_name)
            html_responsive_resource = resources.get_local_html_responsive_resource(
                width
            )
            return html_responsive_resource
        else:
            resources = ImageManagement(file_name)
            html_resource = resources.get_local_html_resource(width)
            return html_resource

    def video_html(self, file_name, width):
        """
        returns an video html tag from a file_name
        """
        resources = VideoManagement(file_name)
        html_resource = resources.get_local_html_resource(width)
        return html_resource

    def image_html_and_text(self, file_name, text, width=60):
        """
        returns an image html tag from a file_name and includes the possibility of add text
        """
        img = self.image_html(file_name, width)
        return img + f"""<br>{text}"""

    def formatting_readable_strings(self, input):
        """
        uses the str.replace() to avoid the underscore in names
        """
        changed_str = input.replace("_", " ")
        return changed_str

    def formatting_strings(self, input):
        """
        uses the str.replace() to returns the underscore in names
        """
        changed_str = input.replace(" ", "_")
        return changed_str

    def _donwload_project_data(self) -> None:
        """Donwloads the project model snapshot from Firestore and stores it
        in the AppState singleton."""
        db = firestore_controller.get_authenticated_client(
            self.app_state.authentication_info.role_service_account_credentials
        )
        self.app_state.project_model = firestore_controller.get_current_project_model(
            db, self._tenant_id
        )



import logging
from txp.ml.training_service.dataset_fetchers.bigquery_dataset_fetcher import (
    BigqueryDatasetFetcher,
)
import ray
from txp.cloud.common import utils as cu
from txp.ml.training_service.model_trainers.basic_model_trainer import BasicModelTrainer
import numpy as np
import pandas as pd
from txp.cloud.controllers import bigquery as bq
from txp.ml.common.tasks.slic_cv_tasks import (
    SlicPatchImageProcessor,
)
from txp.common.ml import models
from txp.ml.training_service import training_exceptions as te
from txp.common.ml.tasks import AssetTask


class ComputerVisionSlicClassificationTrainer(BasicModelTrainer):
    """This trainer will take care of training any of the
    basic SLIC models.
    """

    def __init__(
            self,
            credentials_str,
            models_bucket_name,
            data_warehouse_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level=logging.INFO,
    ):
        logging.basicConfig(level=log_level)
        super(ComputerVisionSlicClassificationTrainer, self).__init__(
            credentials_str, models_bucket_name, prediction_service_topic, models_registry_collection_name, log_level
        )
        self._data_fetcher: BigqueryDatasetFetcher = BigqueryDatasetFetcher(
            credentials_str, data_warehouse_bucket_name, log_level
        )

        self.fetcher = BigqueryDatasetFetcher(
            credentials_str, data_warehouse_bucket_name, log_level
        )

    def _train_task(
            self,
            task,
            task_definition: AssetTask,
            dataset_name,
            dataset_versions,
            tenant_id,
            machine_id,
            task_id,
    ):
        """Trains the received Task Actor Handle which refers to a concrete
        Slic Computer vision recognition implementation.

        If hyper parameters are present in task_definition, they will be passed
        across the common interface for SLIC classes. If not, then defaults
        are used by each implementation.

        Note:  This implementation currently supports only datasets produced directly
            from BigQuery time records. Eventually we can supports CV datasets from
            another sources like HuggingFace or Segments.ai.

        Args:
            task (ActorHandle): object reference to the ray actor process.

            task_definition (Dict): A key value map with possible hyper parameters
                declared for the task.

            dataset_name: The dataset name in the models repository.

            dataset_versions: The list of labeling version that composes
                the dataset.

            tenant_id: The tenant ID to which this training belongs.

            machine_id: The asset to which this trained task belongs.

            task_id: The task identifier (task class name).
        """
        logging.info(
            f"{self.__class__.__name__} starts training for task/tenant: {task_id}/{tenant_id}"
        )
        logging.info(
            f"SLIC task parameters description: {task.task_definition.parameters}"
        )
        gcs_task_path = cu.get_gcs_task_bucket_path(tenant_id, machine_id, task_id)
        time_df = self.fetcher.get_time_dataframe_from_gcs(
            dataset_name, dataset_versions, gcs_task_path
        )

        #  Currently, take only the normal images from the dataset
        images_signals_df = time_df[time_df["perception_name"] == "Image"]

        images_signals_df.loc[:, 'label'] = images_signals_df.label.apply(
            lambda x: self.get_label_value(task_definition, x)
        )

        # join the images
        images_signals_grouped = images_signals_df.groupby("signal_timestamp")

        logging.info("Successfully images loaded dataframe")
        logging.info("Successfully applied label value to images dataframe")

        images_df = []
        proc = SlicPatchImageProcessor(task.encoder._number_of_segments)

        for signal_timestamp, signal_df in images_signals_grouped:
            all_chunks = signal_df.to_dict("records")
            signal = cu.merge_signal_chunks(all_chunks)
            image = bq.get_image_from_raw_data(signal["data"], lower_quality=False)

            if image is not None:
                try:
                    images_df.append(
                        {
                            "image": proc.process_image_patches(
                                np.array(image),
                                task.encoder._patch_x_y_image[0],
                                task.encoder._patch_x_y_image[1],
                            ),
                            "label": signal_df.iloc[0]["label"],
                        }
                    )
                except OSError as e:
                    logging.warning(f"Truncated image file error while processing image sample")
                except Exception as e:
                    logging.warning(f"Unknown error while processing image file sample: {e}")

        df = pd.DataFrame(images_df, columns=["image", "label"])

        encoder_args = {"dataset": df}

        try:
            train_dataset = task.encoder.build_training_dataset({}, "label", **encoder_args)
            # TODO: Task should receive the test_size too!
            task.policy.train(train_dataset, test_size=0.05)
            storage_task = task.save()
            logging.info(f"Task successfully trained with result: {task.policy.accuracy}")
        except:
            raise te.ErrorWhileTrainingModel
        return storage_task

    def get_model_registry(self, task, tenant_id, machine_id, task_id, file_path) -> models.ModelRegistry:
        state = models.ModelState(
            value=models.ModelStateValue.TRAINING,
            creation_date=cu.get_current_utc_datetime_str(),
            publishment_date="",
            deprecation_date=""
        )
        metadata = models.ModelMetadata(
            feedback=models.DefaultModelFeedback(),
            tenant_id=tenant_id,
            machine_id=machine_id,
            task_id=task_id
        )
        registry = models.ModelRegistry(
            metadata=metadata,
            file_path=file_path,
            state=state
        )
        return registry

    def get_model_feedback(self, task) -> models.ModelFeedback:
        return models.SuccessModelFeedback(metrics={"accuracy": task.policy.accuracy})


@ray.remote
class RayComputerVisionSlicClassificationTrainer(
    ComputerVisionSlicClassificationTrainer
):
    def __init__(
            self,
            credentials_str,
            models_bucket_name,
            data_warehouse_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level=logging.INFO,
    ):
        super().__init__(
            credentials_str,
            models_bucket_name,
            data_warehouse_bucket_name,
            prediction_service_topic,
            models_registry_collection_name,
            log_level,
        )



"""The Gateway Package queue.
This module contains the classes that:
    - Support the PackageQueue used to receive Driver collected packages.
"""


# ============================ imports =========================================
import os
from typing import Union
from multiprocessing import Queue
from multiprocessing.managers import BaseProxy
from queuelib import FifoDiskQueue
import psutil
import pickle
from txp.devices.package import GatewayPackage
from txp.common.config import settings
import concurrent.futures
import shutil
import threading
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# Definition of The Package Queue
# ==============================================================================


class PackageQueue:
    """Provides operations to receive packages produced by drivers.
    This PackageQueue is designed to be safely shared across Gateway and Drivers.
    Drivers will put packages in this queue concurrently.
    """

    def __init__(self):
        if (settings.gateway.queue_clean_old_spilled_packages
                and os.path.exists(settings.gateway.queue_storage_path)):
            shutil.rmtree(settings.gateway.queue_storage_path)

        self.pool_executor = concurrent.futures.ThreadPoolExecutor()
        self.received_packages: "Queue[GatewayPackage]" = Queue()

        try:
            self.disk_received_packages = FifoDiskQueue(settings.gateway.queue_storage_path, chunksize=1)
        except FileNotFoundError as ferror:
            log.error("Trying to open Package Queue chunk, but the file was not found. Queue disk storage corrupted.")
            self._backup_corrupted_files()
            self.disk_received_packages = FifoDiskQueue(settings.gateway.queue_storage_path, chunksize=1)
            log.info("New packages Queue created successfully after recovering from corrupted files.")

        self._q_size: int = 0
        """_q_size: the size of the Queue. In MacOS using the provided qsize() may raise
        NotImplementError https://docs.python.org/3/library/multiprocessing.html#multiprocessing.Queue.qsize"""

        recovered_packages = 0
        if len(self.disk_received_packages):
            self._q_size = len(self.disk_received_packages)
            recovered_packages = self._pop_spilled_packages()

        self._disk_push_lock: threading.Lock = threading.Lock()
        """_disk_push_lock: Thread safe for the disk access."""
        self._memory_tolerance: int = settings.gateway.queue_memory_size
        """self._memory_tolerance: indicates the memory percentage at which the packages should be spilled."""
        log.info(f"Packages queue created. Stats on creation: \n"
                 f"Number of spilled packages in disk at initialization: {self._q_size}\n"
                 f"RAM tolerance configured to: {self._memory_tolerance}%\n"
                 f"Number of packages loaded from disk to memory: {recovered_packages}\n"
                 f"Current number of packages in disk: {len(self.disk_received_packages)}\n")

    def _backup_corrupted_files(self) -> None:
        """Performs a recursive copy of the disk queue storage folder, to allow clean instantiation
        of a new FIFO disk queue"""
        import datetime
        current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        backup_folder_suffix = f'_{current_time}'
        shutil.copytree(
            src=settings.gateway.queue_storage_path,
            dst=settings.gateway.queue_storage_path + backup_folder_suffix
        )
        shutil.rmtree(settings.gateway.queue_storage_path)


    def save_current_disk_state(self) -> None:
        """Call this method to save the current disk state of the packages that are currently
        spilled to disk.

        This guarantees that the packages will be available if the Gateway process ends for some
        reason.
        """
        if not len(self.disk_received_packages):
            log.info("Packages Queue was requested to persist spilled packages, but there are "
                     "not packages on disk.")
            return

        self.disk_received_packages.close()
        self.disk_received_packages = FifoDiskQueue(settings.gateway.queue_storage_path, chunksize=1)
        self._create_disk_queue_bak()
        log.info(f"Packages Queue successfully saved {len(self.disk_received_packages)} spilled packages")

    def _create_disk_queue_bak(self) -> None:
        """This method is called when the disk queue is closed with the intent to
        save the packages after process shutdown. It creates a backup of the
        queue file path."""
        path = self.disk_received_packages._infopath()
        copy_path = path + '.bak'
        shutil.copyfile(path, copy_path)  # if copy_path exists, it'll be replaced

    def _pop_spilled_packages(self) -> int:
        """This method is called when spilled packages were found on disk at __init__ time.

        If there are spilled packages, those should be loaded to memory if there's RAM available.

        Returns:
            returns the number of spilled packages loaded into memory.
        """
        c = 0
        while True:
            if (psutil.virtual_memory().percent <= settings.gateway.queue_memory_size
                    and len(self.disk_received_packages)):
                package = self.disk_received_packages.pop()
                package = pickle.loads(package)
                self.received_packages.put(package)
                self._q_size += 1
                c += 1
            else:
                break
        log.debug(f"{c} spilled packages moved from disk to memory")
        self.disk_received_packages.close()
        self.disk_received_packages = FifoDiskQueue(settings.gateway.queue_storage_path, chunksize=1)
        return c

    def close(self):
        # TODO: This close class sometimes corrupts the file after CTRL+C
        #  The backup file created with _create_disk_queue_bak allows a recovery strategy.
        self.disk_received_packages.close()

    def _spill_package(self, package: GatewayPackage):
        encoded_package = pickle.dumps(package)
        self._disk_push_lock.acquire()
        self.disk_received_packages.push(encoded_package)
        self._disk_push_lock.release()
        self._q_size = self._q_size + 1

    def _load_package(self):
        next_package = self.disk_received_packages.pop()
        next_package = pickle.loads(next_package)
        self.received_packages.put(next_package)

    def add_package(self, package: GatewayPackage):
        """Adds the package to this queue.
        Args:
            package: The package sent by the Driver.
        """
        if isinstance(package, GatewayPackage):
            if psutil.virtual_memory().percent >= settings.gateway.queue_memory_size \
                    or len(self.disk_received_packages):
                log.debug(f"Adding GatewayPackage {package.metadata.package_id} in disk storage queue")
                self.pool_executor.submit(self._spill_package, package)
            else:
                log.debug(f"Adding GatewayPackage {package.metadata.package_id} in memory queue")
                self.received_packages.put(package)
                self._q_size = self._q_size + 1

            log.debug(f"Queue Size: {len(self.disk_received_packages)}")

    def get_package(self) -> Union[GatewayPackage, None]:
        """Gets one package from this queue
        Returns:
            A Package if there are at least one package in the queue.
            None otherwise.
        """
        log.debug("Retrieving package from Packages Queue")
        if len(self.disk_received_packages):
            if self.received_packages.empty():
                log.debug("No packages in-memory, retrieving package from disk queue")
                next_package = self.disk_received_packages.pop()
                next_package = pickle.loads(next_package)
                self._q_size = self._q_size - 1
                return next_package
            else:
                log.debug("Retrieving package from in-memory storage and "
                          "concurrently taking 1 package from disk storage to memory")
                self.pool_executor.submit(self._load_package)
        if not self.received_packages.empty():
            log.debug("Retrieving package from in-memory storage. No packages in disk found.")
            self._q_size = self._q_size - 1
            return self.received_packages.get_nowait()
        return None

    def is_empty(self) -> bool:
        return self._q_size == 0

    def size(self) -> int:
        return self._q_size

    def clear(self) -> None:
        log.info("Clearing Packages Queue")
        del self.received_packages
        self.received_packages = Queue()
        self.disk_received_packages.close()
        if os.path.exists(settings.gateway.queue_storage_path):
            shutil.rmtree(settings.gateway.queue_storage_path)
        self.disk_received_packages = FifoDiskQueue(settings.gateway.queue_storage_path, chunksize=1)
        self._q_size = 0


class PackageQueueProxy(BaseProxy):
    """Custom Proxy for a PackageQueue.
    This custom proxy helps with type hints and autocompletion.
    """

    _exposed_ = ["add_package", "get_package", "is_empty", "size", "close", "save_current_disk_state", "clear"]

    def add_package(self, package: GatewayPackage):
        self._callmethod("add_package", (package,))

    def get_package(self) -> Union[GatewayPackage, None]:
        return self._callmethod("get_package")

    def is_empty(self):
        return self._callmethod("is_empty")

    def size(self):
        return self._callmethod("size")

    def clear(self):
        self._callmethod("clear")


    def close(self):
        return self._callmethod("close")

    def save_current_disk_state(self):
        return self._callmethod("save_current_disk_state")



import common.messages_utils
import common.iCOMOX_messages
import struct
import datetime
import communications.Dongle_Communication
from common import helpers

cCLIENT_TYPE_USB    = 0
cCLIENT_TYPE_SMIP   = 1
cCLIENT_TYPE_TCPIP  = 2

cCLIENT_TYPE_BITMASK_USB    = 1 << cCLIENT_TYPE_USB
cCLIENT_TYPE_BITMASK_SMIP   = 1 << cCLIENT_TYPE_SMIP
cCLIENT_TYPE_BITMASK_TCPIP  = 1 << cCLIENT_TYPE_TCPIP

cCLIENT_TYPE_BITMASK_ALL = cCLIENT_TYPE_BITMASK_USB | cCLIENT_TYPE_BITMASK_SMIP | cCLIENT_TYPE_BITMASK_TCPIP

def list_to_first_element(list):
    if len(list) == 0:
        return None
    else:
        return list[0]

class sCOMOX():
    def __init__(self, Type):
        self.Type = Type    # Type of connection (cCLIENT_TYPE_xxx). It is not the type of the iCOMOX!!

        self.Hello = None
        self.Reports = [None] * 256
        # self.Configuration = bytearray()
        self.ConnectionTime = None  # datetime.datetime.now()
        # self.HelloReceivedTime = None

    def same_iCOMOX(self, iComox):
        if iComox.Type != self.Type:
            return False
        if (self.Hello is not None) and (iComox.Hello is not None):
            return self.UniqueID() == iComox.UniqueID()
        return None # Can't determine

    def copy_from(self, iComox):
        if (iComox is None) or (iComox.Type != self.Type):
            raise Exception("sCOMOX.copy_from(): Invalid iComox type")
        # self.Hello = iComox.Hello
        # self.Reports = iComox.Reports
        # self.ConnectionTime = iComox.ConnectionTime

    def clear(self):  # it does not clean the UniqueID field
        self.Type = None
        self.Hello = None
        self.Reports = [None] * 256
        self.ConnectionTime = None

    def UniqueID(self):
        if self.Hello is not None:
            return self.Hello[4:20]
        else:
            return None

    def UniqueIDString(self):
        if self.Hello is not None:
            return helpers.u8s_to_str(arr= self.Hello[4:20], separator="")
        else:
            return None

    def board_type(self):
        if self.Hello is not None:
            return self.Hello[1]
        else:
            return None

    def Name(self):
        if self.Hello is not None:
            return self.Hello[96:128]
        else:
            return None


class sCOMOX_USB(sCOMOX):
    def __init__(self, CommPort):
        sCOMOX.__init__(self, Type=cCLIENT_TYPE_USB)
        self.CommPort = CommPort
        # self.HelloReceivedTime = datetime.datetime.now()

    def clear(self):
        sCOMOX.clear(self)
        self.CommPort = None

    def same_iCOMOX(self, iComox):
        same = sCOMOX.same_iCOMOX(self, iComox=iComox)
        if same is None:
            return self.CommPort == iComox.CommPort
        else:
            return same


class sCOMOX_SMIP(sCOMOX):
    def __init__(self, macAddress, moteID, state, isRouting):
        sCOMOX.__init__(self, Type=cCLIENT_TYPE_SMIP)
        self.macAddress = macAddress
        self.moteID = moteID
        self.state = state  # Dongle_Communication.MOTE_STATE_LOST
        self.isRouting = isRouting
        self.packetID = 0
        self.expected_msg_size = 0
        self.accumulated_msg = bytearray()
        self.HelloRequestTime = None

    def clear(self):
        sCOMOX.clear(self)
        self.macAddress = None
        self.moteID = None

    def same_iCOMOX(self, iComox):
        same = sCOMOX.same_iCOMOX(self, iComox=iComox)
        if same is None:
            return self.macAddress == iComox.macAddress
        else:
            return same

    def copy_from(self, iComox):
        sCOMOX.copy_from(self, iComox=iComox)
        self.moteID = iComox.moteID
        self.state = iComox.state
        self.isRouting = iComox.isRouting


class sCOMOX_TCPIP(sCOMOX):
    def __init__(self, remoteAddress, remotePort, socket):
        sCOMOX.__init__(self, Type=cCLIENT_TYPE_TCPIP)
        self.socket = socket
        self.remoteAddress = remoteAddress
        self.remotePort = remotePort
        self.transmit_buffer = bytearray()
        self.receive_buffer = bytearray()
        self.accumulated_msg = bytearray()

    def clear(self):
        sCOMOX.clear(self)
        self.socket = None
        self.remoteAddress = None
        self.remotePort = None
        self.transmit_buffer = None
        self.receive_buffer = None
        self.accumulated_msg = None

    def copy_from(self, iComox):
        sCOMOX.copy_from(self, iComox=iComox)
        self.socket = iComox.socket
        self.remoteAddress = iComox.remoteAddress
        self.remotePort = iComox.remotePort
        self.transmit_buffer = None
        self.receive_buffer = None
        self.accumulated_msg = None

class sCOMOXs_list():
    def __init__(self):
        self.list = []
        self.current = None

    def add(self, Type, CommPort=None, macAddress=None, moteID=None, state=None, isRouting=None, remoteAddress=None, remotePort=None, socket=None):
        if Type == cCLIENT_TYPE_USB:
            if CommPort is None:
                raise Exception("sCOMOXs_list.add(): Invalid CommPort")
            iComox = sCOMOX_USB(CommPort=CommPort)

        elif Type == cCLIENT_TYPE_SMIP:
            if macAddress is None:
                raise Exception("sCOMOXs_list.add(): Invalid macAddress")
            if moteID is None:
                raise Exception("sCOMOXs_list.add(): Invalid moteID")
            if state is None:
                raise Exception("sCOMOX_list.add(): Invalid state")
            if isRouting is None:
                raise Exception("sCOMOX_list.add(): Invalid isRouting")
            iComox = sCOMOX_SMIP(macAddress=macAddress, moteID=moteID, state=state, isRouting=isRouting)

        elif Type == cCLIENT_TYPE_TCPIP:
            if remoteAddress is None:
                raise Exception("sCOMOXs_list.add(): Invalid remoteAddress")
            if remotePort is None:
                raise Exception("sCOMOXs_list.add(): Invalid remotePort")
            if socket is None:
                raise Exception("sCOMOXs_list.add(): Invalid socket")
            iComox = sCOMOX_TCPIP(remoteAddress=remoteAddress, remotePort=remotePort, socket=socket)

        else:
            raise Exception("sCOMOXs_list.add(): Invalid Type")
        iComox.Type = Type

        ExistingComox = self.exists(iComox=iComox)
        if ExistingComox is not None:
            ExistingComox.copy_from(iComox=iComox)
            iComox.clear()
            del iComox
            return ExistingComox
        else:
            self.list.append(iComox)
            return iComox

    def exists(self, iComox):
        return list_to_first_element(list=[*filter(lambda Cmx : Cmx.same_iCOMOX(iComox=iComox), self.list)])

    def clear(self):
        for iComox in self.list:
            iComox.clear()
        self.list.clear()
        self.current = None

    def all_iComox_of_type(self, Type):
        return [*filter(lambda iCmx : iCmx.Type == Type, self.list)]

    def delete_all_of_type(self, Type):
        list_of_index_to_remove = [*filter(lambda i : self.list[i].Type == Type, range(0, len(self.list)))]
        for i in range(len(list_of_index_to_remove)-1, -1, -1):
            iCmxIndex = list_of_index_to_remove[i]
            self.list[iCmxIndex].clear()
            del self.list[iCmxIndex]

    def is_current_iCOMOX(self, iComox):
        return self.current == iComox

    # def get_all_iCOMOXs_by_type(self, Type):
    #     return [*filter(lambda iComox: iComox.Type == Type, self.list)]

    def find_by_macAddress(self, macAddress):
        return list_to_first_element(list=[*filter(lambda iComox : (iComox.Type == cCLIENT_TYPE_SMIP) and (iComox.macAddress == macAddress), self.list)])

    def find_index_by_macAddress(self, macAddress):
        for i in range(0, len(self.list)):
            iComox = self.list[i]
            if (iComox.Type == cCLIENT_TYPE_SMIP) and (iComox.macAddress == macAddress):
                return i
        return -1

    def find_by_CommPort(self, CommPort):
        return list_to_first_element(list=[*filter(lambda iComox : (iComox.Type == cCLIENT_TYPE_USB) and (iComox.CommPort == CommPort), self.list)])

    def find_index_by_CommPort(self, CommPort):
        for i in range(0, len(self.list)):
            iComox = self.list[i]
            if (iComox.Type == cCLIENT_TYPE_USB) and (iComox.CommPort == CommPort):
                return i
        return -1

    def find_by_UniuqeID(self, UniqueID, TypesBitmask=cCLIENT_TYPE_BITMASK_ALL):
        return list_to_first_element(list=[*filter(lambda iComox : ((iComox.Type & TypesBitmask) != 0) and (iComox.UniqueIDString() == UniqueID), self.list)])

    def find_index_by_uniqueID(self, UniqueID, TypesBitmask=cCLIENT_TYPE_BITMASK_ALL):
        for i in range(0, len(self.list)):
            iComox = self.list[i]
            if ((iComox.Type & TypesBitmask) != 0) and (iComox.UniqueID() == UniqueID):
                return i
        return -1

    def find_by_socket(self, socket):
        return list_to_first_element(list=[*filter(lambda iComox : (iComox.Type == cCLIENT_TYPE_TCPIP) and (iComox.socket == socket), self.list)])

    def find_index_by_socket(self, socket):
        for index in range(0, len(self.list)):
            iComox = self.list[index]
            if (iComox.Type == cCLIENT_TYPE_TCPIP) and (iComox.socket == socket):
                return index
        return None



from dynaconf import Dynaconf, Validator
import os
import pathlib

# file current directory used for relative 'configs' folder access.
current_directory = os.path.dirname(os.path.realpath(__file__))

txp_src_directory = pathlib.Path(current_directory).parents[0]

settings = Dynaconf(settings_files=[f"{current_directory}/configs/settings.toml"])



try:
   import IpMgrConnectorSerialClib
   IpMgrConnectorSerial = IpMgrConnectorSerialClib
   print ("Note: using the C implementation of the IpMgrConnectorSerial connector")
except ImportError:
   import IpMgrConnectorSerial



"""
    This script can be used to add task for machine in firestore.

    This is intended for manual operation of the manual_configuration_edit
    collection used for database initialization.
"""
import google.cloud.firestore as firestore

if __name__ == '__main__':
    collection_name = 'manual_configuration_edit'
    document_src = 'laboratorio_showroom_manual_script'
    subcollection_name = 'edges'
    edge_document = 'beer_demo_spot_2'

    db: firestore.Client = firestore.Client()
    document = db.document(f'{collection_name}/{document_src}/{subcollection_name}/{edge_document}')

    document.set({
        "logical_id": "TXP-ARMROBOT-002_31378a021e68517f2e9b45948521ddbd",
        "device_kind": "ArmRobot",
        "device_type": "ON_OFF_DEVICE",
        "parameters": {
            "physical_id": "TXP-ARMROBOT-002",
            "sensors": [
                {
                    "physical_id": "/dev/v4l/by-id/usb-lihappe8_Corp._USB_2.0_Camera-video-index0",
                    "camera_type": "Normal"
                },
                {
                    "physical_id": "/dev/v4l/by-id/usb-GroupGets_PureThermal__fw:v1.3.0__80180014-5119-3038-3532-373600000000-video-index0",
                    "camera_type": "Thermographic"
                }
            ],
            "virtual_device": {
                "position": [
                    47,
                    23,
                    110,
                    99,
                    86,
                    15
                ],
                "virtual_id_hash": "31378a021e68517f2e9b45948521ddbd"
            }
        },
        "perceptions": {
            "Image": {
                "format": "JPEG",
                "resolution": [
                    640,
                    480
                ],
                "mode": 7
            },
            "ThermalImage": {
                "mode": 7,
                "resolution": [
                    640,
                    480
                ],
                "format": "JPEG"
            }
        }
    })



"""
This module defines the helper function to draw considered fatal
errors in the web view. A fatal error will not allow to render
anything else.

An example of fatal error: Production not completed because it was interrupted.
"""
# =====================imports=====================================
import streamlit as st


def render_fatal_error():
    gateway_state = st.session_state.gateway_state
    st.markdown("-------------")
    st.subheader("Error")

    st.error(gateway_state.details.message)



"""
Definitions of classes to support the concept of perception dimensions in the system.

Classes for Signals are provided, with factory methods to facilitate a single way for Drivers to produce
instances of signals.
"""

# ============================ imports =========================================
from abc import ABC
from typing import Tuple, List, Union
from dataclasses import dataclass
import array
import numpy as np
import txp.common.protos.gateway_package_pb2 as package_proto
import enum
from txp.common.config import settings
import logging
import time
from txp.cloud.common.utils import SignalMode

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


@dataclass
class Signal(ABC):
    """This Signal abstracts the concept of sampled perception dimensions by Edge devices.

    All signals will be shaped in a matrix. Technically, that matrix is created as a List of native arrays.
    Telemetry Protobuf are designed to hold a matrix.
    This is flexible enough for the use cases:

        - A Scalar perception dimension can be represented as a matrix of one row and one column value:
            [[1]]

        - A single axis / double axis / triple axis / n axis perception dimension can be represented as:
            single axis: [[a1 a2 a3 a4 ...... an]]
            double axis: [[a1 ... an], [b1 ... bn]]
            ...

        - Perception dimensions in higher dimensions can be re-shaped to a matrix.


    Args:
        samples((List of native arrays)): The matrix containing the samples for the perceptions.

        perception_dimensions(Tuple[int]): The tuple of ints which indicates the dimensions of the
            perception.

        signal_frequency: The sampling frequency of the signal.

        sampling_resolution: The sampling resolution of the each sample.

        timestamp: a timestamp in nanoseconds when the signal was captured.
    """
    samples: Union[List[array.array]]
    perception_dimensions: Tuple
    signal_frequency: float
    sampling_resolution: int
    timestamp: int

    @classmethod
    def perception_name(cls) -> str:
        """Returns the perception signal name: Vibration, Temperature, ..."""
        pass

    @classmethod
    def is_scalar(cls) -> bool:
        """Concrete signals can describe if they implement an scalar signal"""
        pass

    @classmethod
    def mode(cls) -> SignalMode:
        """Signals will have a processing mode associated"""
        pass

    def get_proto(self) -> package_proto.SignalProto:
        def map_samples_to_protos():
            list_of_dimensions_samples_protos = []
            for array_samples in self.samples:
                proto = package_proto.DimensionSignalSamplesProto()
                proto.samples.extend(array_samples.tolist())
                list_of_dimensions_samples_protos.append(proto)

            return list_of_dimensions_samples_protos

        samples_arrays_protos = map_samples_to_protos()
        signal_proto = package_proto.SignalProto(
            perception_name=self.perception_name(),
            sampling_resolution=self.sampling_resolution,
            signal_frequency=self.signal_frequency,
            timestamp=self.timestamp
        )
        signal_proto.samples.extend(samples_arrays_protos)
        signal_proto.perception_dimensions.extend(list(self.perception_dimensions))

        return signal_proto

    @classmethod
    def build_from_proto(cls, proto: package_proto.SignalProto) -> "Signal":
        if isinstance(proto, package_proto.SignalProto):
            list_of_dimension_signals: List[array.array] = []
            arrays_of_samples = list(proto.samples)
            for dimension_samples in arrays_of_samples:
                l = [float(sample) for sample in dimension_samples.samples]
                list_of_dimension_signals.append(array.array('f', l))

            signal = cls(
                list_of_dimension_signals,
                tuple(proto.perception_dimensions),
                proto.signal_frequency,
                proto.sampling_resolution,
                proto.timestamp
            )

            return signal

        else:
            raise ValueError(
                "Trying to build GatewayConfig from proto, but the proto instance is unknown"
            )

    @classmethod
    def build_signal_from_lists(
            cls,
            samples: List[List],
            perception_dimensions: Tuple,
            signal_frequency: float,
            sampling_resolution: int,
    ) -> "Signal":
        """Builds a signal based on the matrix built as a List of Lists.

        Args:
            samples((List of native arrays)): The matrix containing the samples for the perceptions.
            perception_dimensions(Tuple[int]): The tuple of ints which indicates the dimensions of the
                perception.
            signal_frequency: The sampling frequency of the signal.
            sampling_resolution: The sampling resolution of the each sample.

        Returns:
            Returns the concrete Signal class instance.
        """
        list_of_array_samples: List = [
            array.array("f", samples[i]) for i in range(perception_dimensions[0])
        ]
        timestamp = time.time_ns()
        return cls(
            list_of_array_samples,
            perception_dimensions,
            signal_frequency,
            sampling_resolution,
            timestamp
        )

    @classmethod
    def build_signal_from_numpy(
            cls,
            samples: np.array,
            perception_dimensions: Tuple,
            signal_frequency: float,
            sampling_resolution: int,
    ) -> "Signal":
        """Builds a Signal based on the matrix built as numpy.array

        Args:
            samples((List of native arrays)): The matrix containing the samples for the perceptions.
            perception_dimensions(Tuple[int]): The tuple of ints which indicates the dimensions of the
                perception.
            signal_frequency: The sampling frequency of the signal.
            sampling_resolution: The sampling resolution of the each sample.

        Returns:
            Returns the concrete Signal class instance.
        """
        list_of_array_samples: List = [
            array.array("f", samples[i].tolist())
            for i in range(perception_dimensions[0])
        ]
        timestamp = time.time_ns()
        return cls(
            list_of_array_samples,
            perception_dimensions,
            signal_frequency,
            sampling_resolution,
            timestamp
        )

    @classmethod
    def build_signal_from_value(
            cls, sample: float, signal_frequency: [float, int] = 0.0, sampling_resolution: int = 0
    ) -> "Signal":
        """Build a signal out of an scalar value.

        Args:
            sample (float): the sampled scalar value
            signal_frequency: the frequency of the signal. Preferred to be constant 0 for scalar values.
            sampling_resolution: The sampling resolution. Preferred to be constant 0
                for scalar values. TODO: Check this with team.

        Returns:
            Returns the concrete Signal class instance.
        """
        list_with_sample = [[sample]]
        perception_dimension = (1,)
        return cls.build_signal_from_lists(
            list_with_sample, perception_dimension,
            signal_frequency, sampling_resolution,
        )


# ==============================================================================
# Definition of Concrete Signals classes
# ==============================================================================
class ImageSignal(Signal, ABC):
    """This class provides factory methods to build image signals."""
    @classmethod
    def perception_name(cls) -> str:
        return "Image"

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.IMAGE

    @classmethod
    def build_signal_from_lists(
            cls,
            image: List[List],
            perception_dimensions: Tuple,
            signal_frequency: float,
            sampling_resolution: int,
    ):
        """Builds an ImageSignal from a Lists of Lists.

        Args:
            image: The list of list of samples representing the image.
            perception_dimensions: The perception dimensions of the Signal.
            signal_frequency: The frequency of the signal.
            sampling_resolution: The sampling resolution.
        Returns:
            The ImageSignal for the given np.array
        """
        return super(ImageSignal, cls).build_signal_from_lists(image, perception_dimensions, signal_frequency,
                                                   sampling_resolution)

    @classmethod
    def build_signal_from_numpy(
            cls,
            image: np.array,
            perception_dimensions: Tuple,
            signal_frequency: float,
            sampling_resolution: int,
    ):
        """Builds an ImageSignal from a numpy array

        Args:
            image: The numpy.array containing the sample values.
            perception_dimensions: The perception dimensions of the signal.
            signal_frequency: The frequency of the signal.
            sampling_resolution: The sampling resolution.
        Returns:
            The ImageSignal for the given np.array
        """
        return super(ImageSignal, cls).build_signal_from_numpy(image, perception_dimensions, signal_frequency,
                                                   sampling_resolution)


class ThermalImageSignal(Signal, ABC):
    """This class provides factory methods to build thermal image signals."""
    @classmethod
    def perception_name(cls) -> str:
        return "ThermalImage"

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.IMAGE

    @classmethod
    def build_signal_from_lists(
            cls,
            image: List[List],
            perception_dimensions: Tuple,
            signal_frequency: float,
            sampling_resolution: int,
    ):
        """Builds an ThermalImageSignal from a Lists of Lists.

        Args:
            image: The list of list of samples representing the image.
            perception_dimensions: The perception dimensions of the Signal.
            signal_frequency: The frequency of the signal.
            sampling_resolution: The sampling resolution.
        Returns:
            The ImageSignal for the given np.array
        """
        return super(ThermalImageSignal, cls).build_signal_from_lists(image, perception_dimensions, signal_frequency,
                                                               sampling_resolution)

    @classmethod
    def build_signal_from_numpy(
            cls,
            image: np.array,
            perception_dimensions: Tuple,
            signal_frequency: float,
            sampling_resolution: int,
    ):
        """Builds an ImageSignal from a numpy array

        Args:
            image: The numpy.array containing the sample values.
            perception_dimensions: The perception dimensions of the signal.
            signal_frequency: The frequency of the signal.
            sampling_resolution: The sampling resolution.
        Returns:
            The ImageSignal for the given np.array
        """
        return super(ImageSignal, cls).build_signal_from_numpy(image, perception_dimensions, signal_frequency,
                                                               sampling_resolution)

class VibrationSpeedSignal(Signal):
    """This encapsulates a Vibration Speed signal sample.

    The recommended way of creating instances of this class is to use
    the provided factory methods from VectorSignal.

    Args:
        samples: The built-in List of built-in arrays containing the samples.
        perception_dimensions: The perception dimensions shape for this signal.
        signal_frequency: The frequecy for the signal.
        sampling_resolution: The sampling resolution for the signal.
    """

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def perception_name(cls) -> str:
        return "VibrationSpeed"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.TIME_FFT_PSD


class VibrationMock(Signal):
    """This encapsulates a Vibration Speed signal sample from the MockDriver.

    The recommended way of creating instances of this class is to use
    the provided factory methods from VectorSignal.

    Args:
        samples: The built-in List of built-in arrays containing the samples.
        perception_dimensions: The perception dimensions shape for this signal.
        signal_frequency: The frequecy for the signal.
        sampling_resolution: The sampling resolution for the signal.
    """

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def perception_name(cls) -> str:
        return "VibrationMock"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.TIME_FFT_PSD


class VibrationAccelerationSignal(Signal):
    """This encapsulates a Vibration Acceleration signal sample.

    The recommended way of creating instances of this class is to use
    the provided factory methods from VectorSignal.

    Args:
        samples: The built-in List of built-in arrays containing the samples.
        perception_dimensions: The perception dimensions shape for this signal.
        signal_frequency: The frequecy for the signal.
        sampling_resolution: The sampling resolution for the signal.
    """

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def perception_name(cls) -> str:
        return "VibrationAcceleration"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.TIME_FFT_PSD


class MagnetometerSignal(Signal):
    """This encapsulates a Magnetometer signal sample.

    The recommended way of creating instances of this class is to use
    the provided factory methods from VectorSignal.

    Args:
        samples: The built-in List of built-in arrays containing the samples.
        perception_dimensions: The perception dimensions shape for this signal.
        signal_frequency: The frequecy for the signal.
        sampling_resolution: The sampling resolution for the signal.
    """

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def perception_name(cls) -> str:
        return "Magnetometer"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.TIME_FFT_PSD


class MicrophoneSignal(Signal):
    """This encapsulates a Microphone signal sample.

    The recommended way of creating instances of this class is to use
    the provided factory methods from VectorSignal.

    Args:
        samples: The built-in List of built-in arrays containing the samples.
        perception_dimensions: The perception dimensions shape for this signal.
        signal_frequency: The frequecy for the signal.
        sampling_resolution: The sampling resolution for the signal.
    """

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def perception_name(cls) -> str:
        return "Microphone"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.TIME_FFT_PSD


class ImageMock(ImageSignal):
    """This encapsulates an Image signal sample from the MockDriver.

    The recommended way of creating instances of this class is to use
    the provided factory methods from VectorSignal.

    Args:
        image: The built-in List of built-in arrays containing the samples.
        perception_dimensions: The perception dimensions shape for this signal.
        signal_frequency: The frequecy for the signal.
        sampling_resolution: The sampling resolution for the signal.
    """

    @classmethod
    def is_scalar(cls) -> bool:
        return False

    @classmethod
    def perception_name(cls) -> str:
        return "ImageMock"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.IMAGE


@dataclass
class TemperatureSignal(Signal):
    """This signal encapsulates a temperature measurement as an scalar sample.

    The recommended way of creating instances of this class is to use
    the provided factory methods from ScalarSignal.

    The shape is expected to be described by the tuple value of (1,).
    The samples are expected to be a list of a single array with only one value, in
        order to be consistent with the others signals.
    The signal_frequency is a constant and the
    """
    @classmethod
    def is_scalar(cls) -> bool:
        return True

    @classmethod
    def perception_name(cls) -> str:
        return "Temperature"

    @classmethod
    def mode(cls) -> SignalMode:
        """Todo: Hardcoded mode. This should be passed in the edge configuration"""
        return SignalMode.TIME


class SignalsTableByName(enum.Enum):
    def __new__(cls, signal_class: Signal):
        obj = object.__new__(cls)
        obj._value_ = signal_class.perception_name()
        obj.signal = signal_class
        return obj

    VIBRATION_SPEED = VibrationSpeedSignal
    VIBRATION_ACCELERATION = VibrationAccelerationSignal
    VIBRATION_MOCK = VibrationMock
    MAGNETOMETER = MagnetometerSignal
    MICROPHONE = MicrophoneSignal
    TEMPERATURE = TemperatureSignal
    IMAGE = ImageSignal
    IMAGEMOCK = ImageMock



#!/usr/bin/python

import struct
import operator
import types

from SmartMeshSDK.ApiDefinition import ApiDefinition
from SmartMeshSDK.ApiException  import CommandError
from SmartMeshSDK.utils         import FormatUtils

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('ByteArraySerializer')
log.setLevel(logging.ERROR)
log.addHandler(NullHandler())

class ByteArraySerializer(object):
    '''
    \ingroup ApiDefinition
    
    \brief Serializer/deserializer for byte arrays.
    '''

    def __init__(self,ApiDef):
        self.ApiDef = ApiDef
    
    def serialize(self,commandArray,fieldsToFill):
        
        # log
        if log.isEnabledFor(logging.DEBUG):
            output  = []
            output += ["serialize ..."]
            output += ["- commandArray:     {0}".format(commandArray)]
            output += ["- fieldsToFill:     {0}".format(fieldsToFill)]
            output  = '\n'.join(output)
            log.debug(output)
        
        # validate input
        if type(commandArray)!=list and type(commandArray)!=tuple:
            raise TypeError("First parameter should be a list or tuple, not "+str(type(commandArray)))
        
        # initialize the output
        byteArray  = []
        
        for cmdCounter in range(len(commandArray)):
        
            # packet payload
            definition = self.ApiDef.getDefinition(
                ApiDefinition.ApiDefinition.COMMAND,
                commandArray[:cmdCounter+1]
            )

            fields = [ApiDefinition.Field(fieldRaw,self.ApiDef.fieldOptions)
                         for fieldRaw in definition['request']]
            
            for field in fields:
                thisFieldByteArray = []
                if field.name in ApiDefinition.ApiDefinition.RESERVED:
                    thisFieldByteArray.append(
                        self.ApiDef.subcommandNameToId(
                            ApiDefinition.ApiDefinition.COMMAND,
                            commandArray[:cmdCounter+1],
                            commandArray[cmdCounter+1]
                        )
                    )
                else:
                    val                          = fieldsToFill[field.name]
                    
                    if   field.format==ApiDefinition.FieldFormats.STRING:
                        thisFieldByteArray      += [ord(car) for car in val]
                    
                    elif field.format==ApiDefinition.FieldFormats.BOOL:
                        thisFieldByteArray.append(val)
                    
                    elif field.format==ApiDefinition.FieldFormats.INT:
                        thisFieldByteArray      += [operator.mod(int(val>>(8*i)), 0x100) for i in range(field.length-1, -1, -1)]
                    
                    elif field.format==ApiDefinition.FieldFormats.INTS:
                        if   field.length==1:
                            temp = struct.pack('>b',int(val))
                        elif field.length==2:
                            temp = struct.pack('>h',int(val))
                        elif field.length==4:
                            temp = struct.pack('>i',int(val))
                        else:
                            raise SystemError('field with format='+field.format+' and length='+str(field.length)+' unsupported.')
                        for i in range(len(temp)):
                            thisFieldByteArray.append(ord(temp[i]))
                    
                    elif field.format==ApiDefinition.FieldFormats.HEXDATA:
                        thisFieldByteArray    += val
                    
                    else:
                        raise SystemError('unknown field format='+field.format)
                    
                    # padding
                    while len(thisFieldByteArray) < int(field.length or 0):
                        thisFieldByteArray  = [0x00]+thisFieldByteArray
                
                byteArray = byteArray+thisFieldByteArray
        
        cmdId = self.ApiDef.nameToId(ApiDefinition.ApiDefinition.COMMAND,commandArray)
        
        if log.isEnabledFor(logging.DEBUG):
            output  = []
            output += ["... serialize into"]
            output += ["- cmdId:            {0}".format(cmdId)]
            output += ["- byteArray:        {0}".format(FormatUtils.formatBuffer(byteArray))]
            output  = '\n'.join(output)
            log.debug(output)
        
        return cmdId,byteArray

    def deserialize(self,type,id,byteArray):
        notRcOk         = False
        returnFields    = {}
        nameArray       = [self.ApiDef.idToName(type,id)]
        index           = 0
        
        # log
        if log.isEnabledFor(logging.DEBUG):
            output  = []
            output += ["deserialize ..."]
            output += ["- type:             {0}".format(type)]
            output += ["- id:               {0}".format(id)]
            output += ["- byteArray:        {0}".format(FormatUtils.formatBuffer(byteArray))]
            output  = '\n'.join(output)
            log.debug(output)
        
        continueParsing  = True
        while continueParsing:
            
            fieldDefs   = self.ApiDef.getResponseFields(type,nameArray)
            
            for fieldDef in fieldDefs:
                
                fieldMissing = False
                
                # isolate the piece of the byteArray corresponding to this field
                if fieldDef.length:
                    # this field has an expected length
                    
                    thisFieldArray = byteArray[index:index+fieldDef.length]
                    index         += fieldDef.length
                    
                    if   len(thisFieldArray)==0:
                        # field missing: allowed
                        fieldMissing   = True
                    elif len(thisFieldArray)<fieldDef.length:
                        # incomplete field: not allowed
                        raise CommandError(
                            CommandError.TOO_FEW_BYTES,
                            "incomplete field {0}".format(fieldDef.name),
                        )
                    
                else:
                    thisFieldArray = byteArray[index:]
                    index          = len(byteArray)
                    
                    if len(thisFieldArray)<1:
                        # too few bytes
                        fieldMissing    = True
                
                # find thisFieldValue
                if fieldMissing:
                    thisFieldValue = None
                else:
                    if   fieldDef.format==ApiDefinition.FieldFormats.STRING:
                        thisFieldValue = ''
                        for byte in thisFieldArray:
                            thisFieldValue += chr(byte)
                    
                    elif fieldDef.format==ApiDefinition.FieldFormats.BOOL:
                        if    len(thisFieldArray)==1 and thisFieldArray[0]==0x00:
                            thisFieldValue = False
                        elif  len(thisFieldArray)==1 and thisFieldArray[0]==0x01:
                            thisFieldValue = True
                        else:
                            raise CommandError(CommandError.VALUE_NOT_IN_OPTIONS,
                                               "field="+fieldDef.name+" value="+str(thisFieldValue))
                    
                    elif fieldDef.format==ApiDefinition.FieldFormats.INT:
                        thisFieldValue = 0
                        for i in range(len(thisFieldArray)):
                            thisFieldValue += thisFieldArray[i]*pow(2,8*(len(thisFieldArray)-i-1))
                    
                    elif fieldDef.format==ApiDefinition.FieldFormats.INTS:
                        tempList = [chr(i) for i in thisFieldArray]
                        tempString = ''.join(tempList)
                        if   len(thisFieldArray)==1:
                            thisFieldValue = bytes(thisFieldArray)
                        elif len(thisFieldArray)==2:
                            (thisFieldValue,) = struct.unpack_from('>h',tempString)
                        elif len(thisFieldArray)==4:
                            (thisFieldValue,) = struct.unpack_from('>i',tempString)
                        else:
                            raise SystemError('field with format='+fieldDef.format+' and length='+str(fieldDef.length)+' unsupported.')
                    
                    elif fieldDef.format==ApiDefinition.FieldFormats.HEXDATA:
                        thisFieldValue = thisFieldArray
                    
                    else:
                        raise SystemError('unknown field format='+fieldDef.format)
                    
                    # make sure thisFieldValue in fieldDef.options
                    if fieldDef.options.validOptions:
                        if thisFieldValue not in fieldDef.options.validOptions:
                            raise CommandError(CommandError.VALUE_NOT_IN_OPTIONS,
                                               "field="+fieldDef.name+" value="+str(thisFieldValue))
                
                if fieldDef.name in ApiDefinition.ApiDefinition.RESERVED:
                    # the subcommand specifier cannot be missing
                    if thisFieldValue==None:
                        raise CommandError(
                            CommandError.TOO_FEW_BYTES,
                            "reserved field missing {0}".format(fieldDef.name),
                        )
                    idNextCommand = thisFieldValue
                else:
                    returnFields[fieldDef.name] = thisFieldValue
                    
                # stop if not RC_OK
                if  (
                        (ApiDefinition.ApiDefinition.RC in returnFields) and
                        (
                            returnFields[ApiDefinition.ApiDefinition.RC]!= \
                                ApiDefinition.ApiDefinition.RC_OK
                        )
                   ):
                   notRcOk = True
                   break
            
            # continue if subCommand
            if  (
                    (not notRcOk)           and
                    continueParsing         and
                    self.ApiDef.hasSubcommands(type,nameArray)
                ):
                # find name of subCommand
                nameArray.append(self.ApiDef.subcommandIdToName(type,
                                                                nameArray,
                                                                idNextCommand))
                continueParsing = True
            else:
                continueParsing = False
            
            # stop if not RC_OK
            if  (
                    continueParsing         and
                    notRcOk
                ):
                continueParsing = False
            
            # stop if end of packet reached
            if  (
                    continueParsing         and
                    index>=len(byteArray)
                ):
                continueParsing = False
        
        if log.isEnabledFor(logging.DEBUG):
            output  = []
            output += ["... deserialized into"]
            output += ["- nameArray:        {0}".format(nameArray)]
            output += ["- returnFields:     {0}".format(returnFields)]
            output  = '\n'.join(output)
            log.debug(output)
        
        return nameArray,returnFields
